######################################################################
#
# File: test/unit/b2http/test_b2http.py
#
# Copyright 2019 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import datetime
import locale
import sys
from unittest.mock import MagicMock, call, patch

import apiver_deps
import pytest
import requests
from apiver_deps import USER_AGENT, B2Http, B2HttpApiConfig, ClockSkewHook
from apiver_deps_exception import (
    B2ConnectionError,
    BadDateFormat,
    BadJson,
    BadRequest,
    BrokenPipe,
    ClockSkew,
    ConnectionReset,
    PotentialS3EndpointPassedAsRealm,
    ServiceError,
    TooManyRequests,
    UnknownError,
    UnknownHost,
)

from b2sdk._internal.b2http import setlocale

from ..test_base import TestBase


class TestTranslateErrors:
    def test_ok(self):
        response = MagicMock()
        response.status_code = 200
        actual = B2Http._translate_errors(lambda: response)
        assert response == actual

    def test_partial_content(self):
        response = MagicMock()
        response.status_code = 206
        actual = B2Http._translate_errors(lambda: response)
        assert response == actual

    def test_b2_error(self):
        response = MagicMock()
        response.status_code = 503
        response.content = b'{"status": 503, "code": "server_busy", "message": "busy"}'
        with pytest.raises(ServiceError):
            B2Http._translate_errors(lambda: response)

    def test_broken_pipe(self):
        def fcn():
            raise requests.ConnectionError(
                requests.packages.urllib3.exceptions.ProtocolError(
                    'dummy', OSError(20, 'Broken pipe')
                )
            )

        with pytest.raises(BrokenPipe):
            B2Http._translate_errors(fcn)

    def test_unknown_host(self):
        def fcn():
            raise requests.ConnectionError(
                requests.packages.urllib3.exceptions.MaxRetryError(
                    'AAA nodename nor servname provided, or not known AAA', 'http://example.com'
                )
            )

        with pytest.raises(UnknownHost):
            B2Http._translate_errors(fcn)

    def test_connection_error(self):
        def fcn():
            raise requests.ConnectionError('a message')

        with pytest.raises(B2ConnectionError):
            B2Http._translate_errors(fcn)

    def test_connection_reset(self):
        class SysCallError(Exception):
            pass

        def fcn():
            raise SysCallError('(104, ECONNRESET)')

        with pytest.raises(ConnectionReset):
            B2Http._translate_errors(fcn)

    def test_unknown_error(self):
        def fcn():
            raise Exception('a message')

        with pytest.raises(UnknownError):
            B2Http._translate_errors(fcn)

    def test_too_many_requests(self):
        response = MagicMock()
        response.status_code = 429
        response.headers = {'retry-after': 1}
        response.content = (
            b'{"status": 429, "code": "Too Many requests", "message": "retry after some time"}'
        )
        with pytest.raises(TooManyRequests):
            B2Http._translate_errors(lambda: response)

    def test_invalid_json(self):
        response = MagicMock()
        response.status_code = 400
        response.content = b'{' * 500
        response.url = 'https://example.com'

        with pytest.raises(BadRequest) as exc_info:
            B2Http._translate_errors(lambda: response)

        assert str(exc_info.value) == f'{response.content.decode()} (non_json_response)'

    def test_potential_s3_endpoint_passed_as_realm(self):
        response = MagicMock()
        response.status_code = 400
        response.content = b'<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
        response.url = 'https://s3.us-west-000.backblazeb2.com'

        with pytest.raises(PotentialS3EndpointPassedAsRealm):
            B2Http._translate_errors(lambda: response)

    @pytest.mark.apiver(to_ver=2)
    def test_bucket_id_not_found(self):
        from b2sdk.v2.exception import BucketIdNotFound, v3BucketIdNotFound

        def fcn():
            raise v3BucketIdNotFound('bucket_id')

        with pytest.raises(BucketIdNotFound) as exc_info:
            B2Http._translate_errors(fcn)

        assert str(exc_info.value) == 'Bucket with id=bucket_id not found (bad_bucket_id)'


def test_b2_error__nginx_html():
    """
    While errors with HTML description should not happen, we should not crash on them.
    """
    response = MagicMock()
    response.status_code = 502
    response.content = b'<html><body><h1>502 Bad Gateway</h1></body></html>'
    with pytest.raises(ServiceError) as exc_info:
        B2Http._translate_errors(lambda: response)
    assert response.content.decode('utf-8') in str(exc_info.value)


def test_b2_error__invalid_error_format():
    """
    Handling of invalid error format.

    If server returns valid JSON, but not matching B2 error schema, we should still raise ServiceError.
    """
    response = MagicMock()
    response.status_code = 503
    # valid JSON, but not a valid B2 error (it should be a dict, not a list)
    response.content = b'[]'
    with pytest.raises(ServiceError) as exc_info:
        B2Http._translate_errors(lambda: response)
    assert '503' in str(exc_info.value)


def test_b2_error__invalid_error_values():
    """
    Handling of invalid error values.

    If server returns valid JSON, but not matching B2 error schema, we should still raise ServiceError.
    """
    response = MagicMock()
    response.status_code = 503
    # valid JSON, but not a valid B2 error (code and status values (and therefore types!) are swapped)
    response.content = b'{"code": 503, "message": "Service temporarily unavailable", "status": "service_unavailable"}'
    with pytest.raises(ServiceError) as exc_info:
        B2Http._translate_errors(lambda: response)
    assert '503 Service temporarily unavailable' in str(exc_info.value)


class TestTranslateAndRetry(TestBase):
    def setUp(self):
        self.response = MagicMock()
        self.response.status_code = 200

    def test_works_first_try(self):
        fcn = MagicMock()
        fcn.side_effect = [self.response]
        self.assertIs(self.response, B2Http._translate_and_retry(fcn, 3))

    def test_non_retryable(self):
        with patch('time.sleep') as mock_time:
            fcn = MagicMock()
            fcn.side_effect = [BadJson('a'), self.response]
            with self.assertRaises(BadJson):
                B2Http._translate_and_retry(fcn, 3)
            self.assertEqual([], mock_time.mock_calls)

    def test_works_second_try(self):
        with patch('time.sleep') as mock_time:
            fcn = MagicMock()
            fcn.side_effect = [ServiceError('a'), self.response]
            self.assertIs(self.response, B2Http._translate_and_retry(fcn, 3))
            self.assertEqual([call(1.0)], mock_time.mock_calls)

    def test_never_works(self):
        with patch('time.sleep') as mock_time:
            fcn = MagicMock()
            fcn.side_effect = [
                ServiceError('a'),
                ServiceError('a'),
                ServiceError('a'),
                self.response,
            ]
            with self.assertRaises(ServiceError):
                B2Http._translate_and_retry(fcn, 3)
            self.assertEqual([call(1.0), call(1.5)], mock_time.mock_calls)

    def test_too_many_requests_works_after_sleep(self):
        with patch('time.sleep') as mock_time:
            fcn = MagicMock()
            fcn.side_effect = [TooManyRequests(retry_after_seconds=2), self.response]
            self.assertIs(self.response, B2Http._translate_and_retry(fcn, 3))
            self.assertEqual([call(2)], mock_time.mock_calls)

    def test_too_many_requests_failed_after_sleep(self):
        with patch('time.sleep') as mock_time:
            fcn = MagicMock()
            fcn.side_effect = [
                TooManyRequests(retry_after_seconds=2),
                TooManyRequests(retry_after_seconds=5),
            ]
            with self.assertRaises(TooManyRequests):
                B2Http._translate_and_retry(fcn, 2)
            self.assertEqual([call(2)], mock_time.mock_calls)

    def test_too_many_requests_retry_header_combination_one(self):
        # If the first response didn't have a header, second one has, and third one doesn't have, what should happen?

        with patch('time.sleep') as mock_time:
            fcn = MagicMock()
            fcn.side_effect = [
                TooManyRequests(retry_after_seconds=2),
                TooManyRequests(),
                TooManyRequests(retry_after_seconds=2),
                self.response,
            ]
            self.assertIs(self.response, B2Http._translate_and_retry(fcn, 4))
            self.assertEqual([call(2), call(1.5), call(2)], mock_time.mock_calls)

    def test_too_many_requests_retry_header_combination_two(self):
        # If the first response had header, and the second did not, but the third has header again, what should happen?

        with patch('time.sleep') as mock_time:
            fcn = MagicMock()
            fcn.side_effect = [
                TooManyRequests(),
                TooManyRequests(retry_after_seconds=5),
                TooManyRequests(),
                self.response,
            ]
            self.assertIs(self.response, B2Http._translate_and_retry(fcn, 4))
            self.assertEqual([call(1.0), call(5), call(2.25)], mock_time.mock_calls)


class TestB2Http(TestBase):
    URL = 'http://example.com'
    UA_APPEND = None
    HEADERS = dict(my_header='my_value')
    EXPECTED_HEADERS = {'my_header': 'my_value', 'User-Agent': USER_AGENT}
    EXPECTED_JSON_HEADERS = {
        **EXPECTED_HEADERS,
        'Content-Type': 'application/json',
        'Accept': 'application/json',
    }
    PARAMS = dict(fileSize=100)
    PARAMS_JSON_BYTES = b'{"fileSize": 100}'

    def setUp(self):
        self.session = MagicMock()
        self.response = MagicMock()

        requests = MagicMock()
        requests.Session.return_value = self.session

        if apiver_deps.V <= 1:
            self.b2_http = B2Http(
                requests, install_clock_skew_hook=False, user_agent_append=self.UA_APPEND
            )
        else:
            self.b2_http = B2Http(
                B2HttpApiConfig(
                    requests.Session,
                    install_clock_skew_hook=False,
                    user_agent_append=self.UA_APPEND,
                )
            )

    def test_post_json_return_json(self):
        self.session.request.return_value = self.response
        self.response.status_code = 200
        self.response.content = b'{"color": "blue"}'
        response_dict = self.b2_http.post_json_return_json(self.URL, self.HEADERS, self.PARAMS)
        self.assertEqual({'color': 'blue'}, response_dict)
        (pos_args, kw_args) = self.session.request.call_args
        assert pos_args[:2] == ('POST', self.URL)
        assert kw_args['headers'] == self.EXPECTED_JSON_HEADERS
        assert kw_args['data'] == self.PARAMS_JSON_BYTES

    def test_callback(self):
        callback = MagicMock()
        callback.pre_request = MagicMock()
        callback.post_request = MagicMock()
        self.b2_http.add_callback(callback)
        self.session.request.return_value = self.response
        self.response.status_code = 200
        self.response.content = b'{"color": "blue"}'
        self.b2_http.post_json_return_json(self.URL, self.HEADERS, self.PARAMS)
        callback.pre_request.assert_called_with(
            'POST', 'http://example.com', self.EXPECTED_JSON_HEADERS
        )
        callback.post_request.assert_called_with(
            'POST', 'http://example.com', self.EXPECTED_JSON_HEADERS, self.response
        )

    def test_get_content(self):
        self.session.request.return_value = self.response
        self.response.status_code = 200
        with self.b2_http.get_content(self.URL, self.HEADERS) as r:
            self.assertIs(self.response, r)
        self.session.request.assert_called_with(
            'GET',
            self.URL,
            headers=self.EXPECTED_HEADERS,
            data=None,
            params=None,
            stream=True,
            timeout=(B2Http.CONNECTION_TIMEOUT, B2Http.TIMEOUT),
        )
        self.response.close.assert_not_called()  # prevent premature close() on requests.Response

    def test_head_content(self):
        self.session.request.return_value = self.response
        self.response.status_code = 200
        self.response.headers = {'color': 'blue'}

        response = self.b2_http.head_content(self.URL, self.HEADERS)

        self.assertEqual({'color': 'blue'}, response.headers)
        (pos_args, kw_args) = self.session.request.call_args
        assert pos_args[:2] == ('HEAD', self.URL)
        assert kw_args['headers'] == self.EXPECTED_HEADERS


class TestB2HttpUserAgentAppend(TestB2Http):
    UA_APPEND = 'ua_extra_string'
    EXPECTED_HEADERS = {**TestB2Http.EXPECTED_HEADERS, 'User-Agent': f'{USER_AGENT} {UA_APPEND}'}
    EXPECTED_JSON_HEADERS = {
        **TestB2Http.EXPECTED_JSON_HEADERS,
        'User-Agent': EXPECTED_HEADERS['User-Agent'],
    }


class TestSetLocaleContextManager(TestBase):
    def test_set_locale_context_manager(self):
        # C.UTF-8 on Ubuntu 18.04 Bionic, C.utf8 on Ubuntu 22.04 Jammy
        # Neither macOS nor Windows have C.UTF-8 locale, and they use `en_US.UTF-8`.
        # Since Python 3.12, locale.normalize no longer falls back
        # to the `en_US` version, so we're providing it here manually.
        test_locale = locale.normalize('C.UTF-8' if sys.platform == 'linux' else 'en_US.UTF-8')
        other_locale = 'C'

        saved = locale.setlocale(locale.LC_ALL)
        if saved == test_locale:
            test_locale, other_locale = other_locale, test_locale

        locale.setlocale(locale.LC_ALL, other_locale)
        with setlocale(test_locale):
            assert locale.setlocale(category=locale.LC_ALL) == test_locale
        locale.setlocale(locale.LC_ALL, saved)


class TestClockSkewHook(TestBase):
    def test_bad_format(self):
        response = MagicMock()
        response.headers = {'Date': 'bad format'}
        with self.assertRaises(BadDateFormat):
            ClockSkewHook().post_request('POST', 'http://example.com', {}, response)

    def test_bad_month(self):
        response = MagicMock()
        response.headers = {'Date': 'Fri, 16 XXX 2016 20:52:30 GMT'}
        with self.assertRaises(BadDateFormat):
            ClockSkewHook().post_request('POST', 'http://example.com', {}, response)

    def test_no_skew(self):
        now = datetime.datetime.now(datetime.timezone.utc)
        now_str = now.strftime('%a, %d %b %Y %H:%M:%S GMT')
        response = MagicMock()
        response.headers = {'Date': now_str}
        ClockSkewHook().post_request('POST', 'http://example.com', {}, response)

    def test_positive_skew(self):
        now = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(minutes=11)
        now_str = now.strftime('%a, %d %b %Y %H:%M:%S GMT')
        response = MagicMock()
        response.headers = {'Date': now_str}
        with self.assertRaises(ClockSkew):
            ClockSkewHook().post_request('POST', 'http://example.com', {}, response)

    def test_negative_skew(self):
        now = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(minutes=-11)
        now_str = now.strftime('%a, %d %b %Y %H:%M:%S GMT')
        response = MagicMock()
        response.headers = {'Date': now_str}
        with self.assertRaises(ClockSkew):
            ClockSkewHook().post_request('POST', 'http://example.com', {}, response)



######################################################################
#
# File: test/integration/test_sync.py
#
# Copyright 2024 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import io
import time

import pytest

from b2sdk.v2 import (
    CompareVersionMode,
    NewerFileSyncMode,
    Synchronizer,
    SyncReport,
    parse_folder,
)


@pytest.fixture
def local_folder_with_files(tmp_path):
    folder = tmp_path / 'test'
    folder.mkdir()
    (folder / 'a').mkdir()
    (folder / 'a' / 'foo').write_bytes(b'foo')
    # space in the name is important as it influences lexicographical sorting used by B2
    (folder / 'a b').mkdir()
    (folder / 'a b' / 'bar').write_bytes(b'bar')
    return folder


def test_sync_folder(b2_api, local_folder_with_files, b2_subfolder):
    source_folder = parse_folder(str(local_folder_with_files), b2_api)
    dest_folder = parse_folder(b2_subfolder, b2_api)

    synchronizer = Synchronizer(
        max_workers=10,
        newer_file_mode=NewerFileSyncMode.REPLACE,
        compare_version_mode=CompareVersionMode.MODTIME,
        compare_threshold=10,  # ms
    )

    def sync_and_report():
        buf = io.StringIO()
        reporter = SyncReport(buf, no_progress=True)
        with reporter:
            synchronizer.sync_folders(
                source_folder=source_folder,
                dest_folder=dest_folder,
                now_millis=int(1000 * time.time()),
                reporter=reporter,
            )
        return reporter

    report = sync_and_report()
    assert report.total_transfer_files == 2
    assert report.total_transfer_bytes == 6

    second_sync_report = sync_and_report()
    assert second_sync_report.total_transfer_files == 0
    assert second_sync_report.total_transfer_bytes == 0



######################################################################
#
# File: test/unit/utils/test_range_.py
#
# Copyright 2024 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import pytest


def test_range_initialization(apiver_module):
    r = apiver_module.Range(0, 10)
    assert r.start == 0
    assert r.end == 10


def test_range_eq(apiver_module):
    r = apiver_module.Range(5, 10)
    assert r == apiver_module.Range(5, 10)
    assert r != apiver_module.Range(5, 11)
    assert r != apiver_module.Range(6, 10)


def test_range_initialization_invalid(apiver_module):
    with pytest.raises(AssertionError):
        apiver_module.Range(10, 0)


def test_range_from_header(apiver_module):
    r = apiver_module.Range.from_header('bytes=0-11')
    assert r.start == 0
    assert r.end == 11


@pytest.mark.parametrize(
    'raw_range_header, start, end, total_length',
    [
        ('bytes 0-11', 0, 11, None),
        ('bytes 1-11/*', 1, 11, None),
        ('bytes 10-110/200', 10, 110, 200),
    ],
)
def test_range_from_header_with_size(apiver_module, raw_range_header, start, end, total_length):
    r, length = apiver_module.Range.from_header_with_size(raw_range_header)
    assert r.start == start
    assert r.end == end
    assert length == total_length


def test_range_size(apiver_module):
    r = apiver_module.Range(0, 10)
    assert r.size() == 11


def test_range_subrange(apiver_module):
    r = apiver_module.Range(1, 10)
    assert r.subrange(0, 9) == apiver_module.Range(1, 10)
    assert r.subrange(2, 5) == apiver_module.Range(3, 6)


def test_range_subrange_invalid(apiver_module):
    r = apiver_module.Range(0, 10)
    with pytest.raises(AssertionError):
        r.subrange(5, 15)


def test_range_as_tuple(apiver_module):
    r = apiver_module.Range(0, 10)
    assert r.as_tuple() == (0, 10)


def test_range_repr(apiver_module):
    r = apiver_module.Range(0, 10)
    assert repr(r) == 'Range(0, 10)'


def test_empty_range(apiver_module):
    r = apiver_module.EMPTY_RANGE
    assert r.size() == 0



######################################################################
#
# File: test/unit/stream/test_progress.py
#
# Copyright 2024 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import io
from unittest.mock import Mock

from apiver_deps import ReadingStreamWithProgress


def test_reading_stream_with_progress(tmp_path):
    stream = io.BytesIO(b'1234567890')
    progress_listener = Mock()
    with ReadingStreamWithProgress(stream, progress_listener=progress_listener) as wrapped_stream:
        assert wrapped_stream.read(1) == b'1'
        assert wrapped_stream.read(2) == b'23'
        assert wrapped_stream.read(3) == b'456'

        assert progress_listener.bytes_completed.call_count == 3
        assert wrapped_stream.bytes_completed == 6

    assert not stream.closed


def test_reading_stream_with_progress__not_closing_wrapped_stream(tmp_path):
    stream = io.BytesIO(b'1234567890')
    progress_listener = Mock()
    with ReadingStreamWithProgress(stream, progress_listener=progress_listener) as wrapped_stream:
        assert wrapped_stream.read()

    assert not stream.closed


def test_reading_stream_with_progress__closed_proxy(tmp_path):
    """
    Test that the wrapped stream is closed when the original stream is closed.

    This is important for Python 3.13+ to prevent:
    'Exception ignored in: <b2sdk._internal.stream.progress.ReadingStreamWithProgress object at 0x748cf40d5180>'
    messages.
    """
    stream = io.BytesIO(b'1234567890')
    progress_listener = Mock()
    wrapped_stream = ReadingStreamWithProgress(stream, progress_listener=progress_listener)

    assert not stream.closed
    stream.close()
    assert wrapped_stream.closed



######################################################################
#
# File: test/unit/utils/test_escape.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import pytest

from b2sdk._internal.utils.escape import (
    escape_control_chars,
    substitute_control_chars,
    unprintable_to_hex,
)


@pytest.mark.parametrize(
    (
        'input_',
        'expected_unprintable_to_hex',
        'expected_escape_control_chars',
        'expected_substitute_control_chars',
    ),
    [
        ('', '', '', ('', False)),
        (' abc-z', ' abc-z', "' abc-z'", (' abc-z', False)),
        ('a\x7fb', 'a\\x7fb', "'a\\x7fb'", ('a�b', True)),
        ('a\x00b a\x9fb ', 'a\\x00b a\\x9fb ', "'a\\x00b a\\x9fb '", ('a�b a�b ', True)),
        ('a\x7fb\nc', 'a\\x7fb\nc', "'a\\x7fb\nc'", ('a�b\nc', True)),
        ('\x9bT\x9bEtest', '\\x9bT\\x9bEtest', "'\\x9bT\\x9bEtest'", ('�T�Etest', True)),
        (
            '\x1b[32mC\x1b[33mC\x1b[34mI',
            '\\x1b[32mC\\x1b[33mC\\x1b[34mI',
            "'\\x1b[32mC\\x1b[33mC\\x1b[34mI'",
            ('�[32mC�[33mC�[34mI', True),
        ),
    ],
)
def test_unprintable_to_hex(
    input_,
    expected_unprintable_to_hex,
    expected_escape_control_chars,
    expected_substitute_control_chars,
):
    assert unprintable_to_hex(input_) == expected_unprintable_to_hex
    assert escape_control_chars(input_) == expected_escape_control_chars
    assert substitute_control_chars(input_) == expected_substitute_control_chars


def test_unprintable_to_hex__none():
    """
    Test that unprintable_to_hex handles None.

    This was unintentionally supported and is only kept for compatibility.
    """
    assert unprintable_to_hex(None) is None  # type: ignore


def test_escape_control_chars__none():
    """
    Test that escape_control_chars handles None.

    This was unintentionally supported and is only kept for compatibility.
    """
    assert escape_control_chars(None) is None  # type: ignore


def test_substitute_control_chars__none():
    with pytest.raises(TypeError):
        substitute_control_chars(None)  # type: ignore



######################################################################
#
# File: test/integration/test_upload.py
#
# Copyright 2021 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import io

from b2sdk._internal.b2http import B2Http
from b2sdk._internal.encryption.setting import EncryptionKey, EncryptionSetting
from b2sdk._internal.encryption.types import EncryptionAlgorithm, EncryptionMode
from b2sdk.v2 import B2RawHTTPApi

from .base import IntegrationTestBase
from .test_raw_api import authorize_raw_api


class TestUnboundStreamUpload(IntegrationTestBase):
    def assert_data_uploaded_via_stream(self, data: bytes, part_size: int | None = None):
        bucket = self.create_bucket()
        stream = io.BytesIO(data)
        file_name = 'unbound_stream'

        bucket.upload_unbound_stream(stream, file_name, recommended_upload_part_size=part_size)

        downloaded_data = io.BytesIO()
        bucket.download_file_by_name(file_name).save(downloaded_data)

        assert downloaded_data.getvalue() == data

    def test_streamed_small_buffer(self):
        # 20kb
        data = b'a small data content' * 1024
        self.assert_data_uploaded_via_stream(data)

    def test_streamed_large_buffer_small_part_size(self):
        # 10mb
        data = b'a large data content' * 512 * 1024
        # 5mb, the smallest allowed part size
        self.assert_data_uploaded_via_stream(data, part_size=5 * 1024 * 1024)


class TestUploadLargeFile(IntegrationTestBase):
    def test_ssec_key_id(self):
        sse_c = EncryptionSetting(
            mode=EncryptionMode.SSE_C,
            algorithm=EncryptionAlgorithm.AES256,
            key=EncryptionKey(secret=b'********************************', key_id='some-id'),
        )

        raw_api = B2RawHTTPApi(B2Http())

        auth_dict = authorize_raw_api(raw_api)
        account_auth_token = auth_dict['authorizationToken']
        api_url = auth_dict['apiInfo']['storageApi']['apiUrl']
        bucket = self.create_bucket()

        large_info = raw_api.start_large_file(
            api_url,
            account_auth_token,
            bucket.id_,
            'test_largefile_sse_c.txt',
            'text/plain',
            None,
            server_side_encryption=sse_c,
        )

        assert large_info['fileInfo'] == {
            'sse_c_key_id': sse_c.key.key_id,
        }
        assert large_info['serverSideEncryption'] == {
            'algorithm': 'AES256',
            'customerKeyMd5': 'SaaDheEjzuynJH8eW6AEpQ==',
            'mode': 'SSE-C',
        }



######################################################################
#
# File: test/integration/bucket_cleaner.py
#
# Copyright 2022 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import logging

from b2sdk.v2 import (
    NO_RETENTION_FILE_SETTING,
    B2Api,
    Bucket,
    LegalHold,
    RetentionMode,
    current_time_millis,
)
from b2sdk.v2.exception import BadRequest

from .helpers import BUCKET_CREATED_AT_MILLIS, GENERAL_BUCKET_NAME_PREFIX

ONE_HOUR_MILLIS = 60 * 60 * 1000

logger = logging.getLogger(__name__)


class BucketCleaner:
    def __init__(
        self, dont_cleanup_old_buckets: bool, b2_api: B2Api, current_run_prefix: str | None = None
    ):
        self.current_run_prefix = current_run_prefix
        self.dont_cleanup_old_buckets = dont_cleanup_old_buckets
        self.b2_api = b2_api

    def _should_remove_bucket(self, bucket: Bucket):
        if self.current_run_prefix and bucket.name.startswith(self.current_run_prefix):
            return True
        if self.dont_cleanup_old_buckets:
            return False
        if bucket.name.startswith(GENERAL_BUCKET_NAME_PREFIX):
            if BUCKET_CREATED_AT_MILLIS in bucket.bucket_info:
                if (
                    int(bucket.bucket_info[BUCKET_CREATED_AT_MILLIS])
                    < current_time_millis() - ONE_HOUR_MILLIS
                ):
                    return True
        return False

    def cleanup_buckets(self):
        buckets = self.b2_api.list_buckets()
        for bucket in buckets:
            self.cleanup_bucket(bucket)

    def cleanup_bucket(self, bucket: Bucket):
        b2_api = self.b2_api
        if not self._should_remove_bucket(bucket):
            logger.info('Skipping bucket removal:', bucket.name)
        else:
            logger.info('Trying to remove bucket:', bucket.name)
            files_leftover = False
            try:
                b2_api.delete_bucket(bucket)
            except BadRequest:
                logger.info('Bucket is not empty, removing files')
                files_leftover = True

            if files_leftover:
                files_leftover = False
                file_versions = bucket.ls(latest_only=False, recursive=True)
                for file_version_info, _ in file_versions:
                    if file_version_info.file_retention:
                        if file_version_info.file_retention.mode == RetentionMode.GOVERNANCE:
                            logger.info(
                                'Removing retention from file version: %s', file_version_info.id_
                            )
                            b2_api.update_file_retention(
                                file_version_info.id_,
                                file_version_info.file_name,
                                NO_RETENTION_FILE_SETTING,
                                True,
                            )
                        elif file_version_info.file_retention.mode == RetentionMode.COMPLIANCE:
                            if (
                                file_version_info.file_retention.retain_until
                                > current_time_millis()
                            ):
                                logger.info(
                                    'File version: %s cannot be removed due to compliance mode retention',
                                    file_version_info.id_,
                                )
                                files_leftover = True
                                continue
                        elif file_version_info.file_retention.mode == RetentionMode.NONE:
                            pass
                        else:
                            raise ValueError(
                                f'Unknown retention mode: {file_version_info.file_retention.mode}'
                            )
                    if file_version_info.legal_hold.is_on():
                        logger.info(
                            'Removing legal hold from file version: %s', file_version_info.id_
                        )
                        b2_api.update_file_legal_hold(
                            file_version_info.id_, file_version_info.file_name, LegalHold.OFF
                        )
                    logger.info('Removing file version:', file_version_info.id_)
                    b2_api.delete_file_version(file_version_info.id_, file_version_info.file_name)

                if files_leftover:
                    logger.info('Unable to remove bucket because some retained files remain')
                    return
                else:
                    b2_api.delete_bucket(bucket)
            logger.info('Removed bucket:', bucket.name)



######################################################################
#
# File: test/unit/v2/test_utils.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import os
import os.path

import pytest

from .apiver.apiver_deps import TempDir


def test_temp_dir() -> None:
    with pytest.deprecated_call():
        with TempDir() as temp_dir:
            assert os.path.exists(temp_dir)
    assert not os.path.exists(temp_dir)



######################################################################
#
# File: test/unit/utils/test_thread_pool.py
#
# Copyright 2024 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from concurrent.futures import Future

import pytest

from b2sdk._internal.utils.thread_pool import LazyThreadPool


class TestLazyThreadPool:
    @pytest.fixture
    def thread_pool(self):
        return LazyThreadPool()

    def test_submit(self, thread_pool):
        future = thread_pool.submit(sum, (1, 2))
        assert isinstance(future, Future)
        assert future.result() == 3

    def test_set_size(self, thread_pool):
        thread_pool.set_size(10)
        assert thread_pool.get_size() == 10

    def test_get_size(self, thread_pool):
        assert thread_pool.get_size() > 0

    def test_set_size__after_submit(self, thread_pool):
        future = thread_pool.submit(sum, (1, 2))

        thread_pool.set_size(7)
        assert thread_pool.get_size() == 7

        assert future.result() == 3

        assert thread_pool.submit(sum, (1,)).result() == 1



######################################################################
#
# File: test/unit/v_all/test_transfer.py
#
# Copyright 2022 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

from unittest.mock import Mock

from apiver_deps import DownloadManager, UploadManager

from ..test_base import TestBase


class TestDownloadManager(TestBase):
    def test_set_thread_pool_size(self) -> None:
        download_manager = DownloadManager(services=Mock())
        assert download_manager.get_thread_pool_size() > 0

        pool_size = 21
        download_manager.set_thread_pool_size(pool_size)
        assert download_manager.get_thread_pool_size() == pool_size


class TestUploadManager(TestBase):
    def test_set_thread_pool_size(self) -> None:
        upload_manager = UploadManager(services=Mock())
        assert upload_manager.get_thread_pool_size() > 0

        pool_size = 37
        upload_manager.set_thread_pool_size(pool_size)
        assert upload_manager.get_thread_pool_size() == pool_size



######################################################################
#
# File: test/unit/v1/test_utils.py
#
# Copyright 2019 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

from ..test_base import TestBase
from .deps import (
    b2_url_decode,
    b2_url_encode,
    choose_part_ranges,
    format_and_scale_fraction,
    format_and_scale_number,
)

# These are from the B2 Docs (https://www.backblaze.com/b2/docs/string_encoding.html)
ENCODING_TEST_CASES = [
    {'fullyEncoded': '%20', 'minimallyEncoded': '+', 'string': ' '},
    {'fullyEncoded': '%21', 'minimallyEncoded': '!', 'string': '!'},
    {'fullyEncoded': '%22', 'minimallyEncoded': '%22', 'string': '"'},
    {'fullyEncoded': '%23', 'minimallyEncoded': '%23', 'string': '#'},
    {'fullyEncoded': '%24', 'minimallyEncoded': '$', 'string': '$'},
    {'fullyEncoded': '%25', 'minimallyEncoded': '%25', 'string': '%'},
    {'fullyEncoded': '%26', 'minimallyEncoded': '%26', 'string': '&'},
    {'fullyEncoded': '%27', 'minimallyEncoded': "'", 'string': "'"},
    {'fullyEncoded': '%28', 'minimallyEncoded': '(', 'string': '('},
    {'fullyEncoded': '%29', 'minimallyEncoded': ')', 'string': ')'},
    {'fullyEncoded': '%2A', 'minimallyEncoded': '*', 'string': '*'},
    {'fullyEncoded': '%2B', 'minimallyEncoded': '%2B', 'string': '+'},
    {'fullyEncoded': '%2C', 'minimallyEncoded': '%2C', 'string': ','},
    {'fullyEncoded': '%2D', 'minimallyEncoded': '-', 'string': '-'},
    {'fullyEncoded': '%2E', 'minimallyEncoded': '.', 'string': '.'},
    {'fullyEncoded': '/', 'minimallyEncoded': '/', 'string': '/'},
    {'fullyEncoded': '%30', 'minimallyEncoded': '0', 'string': '0'},
    {'fullyEncoded': '%39', 'minimallyEncoded': '9', 'string': '9'},
    {'fullyEncoded': '%3A', 'minimallyEncoded': ':', 'string': ':'},
    {'fullyEncoded': '%3B', 'minimallyEncoded': ';', 'string': ';'},
    {'fullyEncoded': '%3C', 'minimallyEncoded': '%3C', 'string': '<'},
    {'fullyEncoded': '%3D', 'minimallyEncoded': '=', 'string': '='},
    {'fullyEncoded': '%3E', 'minimallyEncoded': '%3E', 'string': '>'},
    {'fullyEncoded': '%3F', 'minimallyEncoded': '%3F', 'string': '?'},
    {'fullyEncoded': '%40', 'minimallyEncoded': '@', 'string': '@'},
    {'fullyEncoded': '%41', 'minimallyEncoded': 'A', 'string': 'A'},
    {'fullyEncoded': '%5A', 'minimallyEncoded': 'Z', 'string': 'Z'},
    {'fullyEncoded': '%5B', 'minimallyEncoded': '%5B', 'string': '['},
    {'fullyEncoded': '%5C', 'minimallyEncoded': '%5C', 'string': '\\'},
    {'fullyEncoded': '%5D', 'minimallyEncoded': '%5D', 'string': ']'},
    {'fullyEncoded': '%5E', 'minimallyEncoded': '%5E', 'string': '^'},
    {'fullyEncoded': '%5F', 'minimallyEncoded': '_', 'string': '_'},
    {'fullyEncoded': '%60', 'minimallyEncoded': '%60', 'string': '`'},
    {'fullyEncoded': '%61', 'minimallyEncoded': 'a', 'string': 'a'},
    {'fullyEncoded': '%7A', 'minimallyEncoded': 'z', 'string': 'z'},
    {'fullyEncoded': '%7B', 'minimallyEncoded': '%7B', 'string': '{'},
    {'fullyEncoded': '%7C', 'minimallyEncoded': '%7C', 'string': '|'},
    {'fullyEncoded': '%7D', 'minimallyEncoded': '%7D', 'string': '}'},
    {'fullyEncoded': '%7E', 'minimallyEncoded': '~', 'string': '~'},
    {'fullyEncoded': '%7F', 'minimallyEncoded': '%7F', 'string': '\u007f'},
    {
        'fullyEncoded': '%E8%87%AA%E7%94%B1',
        'minimallyEncoded': '%E8%87%AA%E7%94%B1',
        'string': '\u81ea\u7531',
    },
    {'fullyEncoded': '%F0%90%90%80', 'minimallyEncoded': '%F0%90%90%80', 'string': '\U00010400'},
]


class TestUrlEncoding(TestBase):
    def test_it(self):
        for test_case in ENCODING_TEST_CASES:
            string = test_case['string']
            fully_encoded = test_case['fullyEncoded']
            minimally_encoded = test_case['minimallyEncoded']
            encoded = b2_url_encode(string)

            expected_encoded = (minimally_encoded, fully_encoded)
            if encoded not in expected_encoded:
                print(f'string: {repr(string)}   encoded: {encoded}   expected: {expected_encoded}')
            self.assertTrue(encoded in expected_encoded)
            self.assertEqual(string, b2_url_decode(fully_encoded))
            self.assertEqual(string, b2_url_decode(minimally_encoded))


class TestChooseParts(TestBase):
    def test_it(self):
        self._check_one([(0, 100), (100, 100)], 200, 100)
        self._check_one([(0, 149), (149, 150)], 299, 100)
        self._check_one([(0, 100), (100, 100), (200, 100)], 300, 100)

        ten_TB = 10 * 1000 * 1000 * 1000 * 1000
        one_GB = 1000 * 1000 * 1000

        expected = [(i * one_GB, one_GB) for i in range(10000)]
        actual = choose_part_ranges(ten_TB, 100 * 1000 * 1000)
        self.assertEqual(expected, actual)

    def _check_one(self, expected, content_length, min_part_size):
        self.assertEqual(expected, choose_part_ranges(content_length, min_part_size))


class TestFormatAndScaleNumber(TestBase):
    def test_it(self):
        self._check_one('1 B', 1)
        self._check_one('999 B', 999)
        self._check_one('1.00 kB', 1000)
        self._check_one('999 kB', 999000)

    def _check_one(self, expected, x):
        self.assertEqual(expected, format_and_scale_number(x, 'B'))


class TestFormatAndScaleFraction(TestBase):
    def test_it(self):
        self._check_one('0 / 100 B', 0, 100)
        self._check_one('0.0 / 10.0 kB', 0, 10000)
        self._check_one('9.4 / 10.0 kB', 9400, 10000)

    def _check_one(self, expected, numerator, denominator):
        self.assertEqual(expected, format_and_scale_fraction(numerator, denominator, 'B'))



######################################################################
#
# File: test/integration/test_file_version_attributes.py
#
# Copyright 2021 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import datetime as dt

from .base import IntegrationTestBase


class TestFileVersionAttributes(IntegrationTestBase):
    def _assert_object_has_attributes(self, object, kwargs):
        for key, value in kwargs.items():
            assert getattr(object, key) == value

    def test_file_info_b2_attributes(self):
        # This test checks that attributes that are internally represented as file_info items with prefix `b2-`
        # are saved and retrieved correctly.

        bucket = self.create_bucket()
        expected_attributes = {
            'cache_control': 'max-age=3600',
            'expires': 'Wed, 21 Oct 2105 07:28:00 GMT',
            'content_disposition': 'attachment; filename="fname.ext"',
            'content_encoding': 'utf-8',
            'content_language': 'en',
        }
        kwargs = {
            **expected_attributes,
            'expires': dt.datetime(2105, 10, 21, 7, 28, tzinfo=dt.timezone.utc),
        }

        file_version = bucket.upload_bytes(b'0', 'file', **kwargs)
        self._assert_object_has_attributes(file_version, expected_attributes)

        file_version = bucket.get_file_info_by_id(file_version.id_)
        self._assert_object_has_attributes(file_version, expected_attributes)

        download_file = bucket.download_file_by_id(file_version.id_)
        self._assert_object_has_attributes(download_file.download_version, expected_attributes)

        copied_version = bucket.copy(
            file_version.id_,
            'file_copy',
            content_type='text/plain',
            **{**kwargs, 'content_language': 'de'},
        )
        self._assert_object_has_attributes(
            copied_version, {**expected_attributes, 'content_language': 'de'}
        )



######################################################################
#
# File: test/unit/internal/transfer/downloader/test_parallel.py
#
# Copyright 2024 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import hashlib
from concurrent.futures import ThreadPoolExecutor
from io import BytesIO
from unittest.mock import Mock

import pytest
from requests import RequestException


def mock_download_response_factory(apiver_module, bucket, file_size: int = 0):
    hasher = hashlib.sha1()

    dummy_data = b'dummy'
    file_content = (dummy_data * (file_size // len(dummy_data) + 1))[:file_size]
    file_version = bucket.upload_bytes(file_content, f'dummy_file_{file_size}.txt')
    hasher.update(file_content)

    url = bucket.api.session.get_download_url_by_name(bucket.name, file_version.file_name)
    response = bucket.api.services.session.download_file_from_url(url).__enter__()

    return response, apiver_module.DownloadVersionFactory(bucket.api).from_response_headers(
        response.headers
    )


@pytest.fixture
def thread_pool():
    with ThreadPoolExecutor(max_workers=10) as executor:
        yield executor


@pytest.fixture
def output_file():
    return BytesIO()


@pytest.fixture
def downloader(apiver_module, thread_pool):
    return apiver_module.ParallelDownloader(
        min_part_size=10,
        force_chunk_size=5,
        thread_pool=thread_pool,
    )


def test_download_empty_file(apiver_module, b2api, bucket, downloader, output_file):
    file_size = 0
    mock_response, download_version = mock_download_response_factory(
        apiver_module, bucket, file_size=file_size
    )
    mock_response.close = Mock(side_effect=mock_response.close)

    bytes_written, hash_hex = downloader.download(
        output_file, mock_response, download_version, b2api.session
    )

    assert bytes_written == file_size
    assert hash_hex == 'da39a3ee5e6b4b0d3255bfef95601890afd80709'
    assert output_file.getvalue() == b''
    mock_response.close.assert_called_once()


def test_download_file(apiver_module, b2api, bucket, downloader, output_file):
    file_size = 100
    mock_response, download_version = mock_download_response_factory(
        apiver_module, bucket, file_size=file_size
    )
    mock_response.close = Mock(side_effect=mock_response.close)

    bytes_written, hash_hex = downloader.download(
        output_file, mock_response, download_version, b2api.session
    )

    assert bytes_written == file_size
    assert hash_hex == '7804df8c623573ccfc1993e04981006e5bc30383'
    assert output_file.getvalue() == b'dummy' * 20
    mock_response.close.assert_called_once()


def test_download_file__data_stream_error__in_first_response(
    apiver_module, b2api, bucket, downloader, output_file
):
    """
    Test that the downloader handles a stream error in the first response.
    """
    file_size = 100
    mock_response, download_version = mock_download_response_factory(
        apiver_module, bucket, file_size=file_size
    )

    def iter_content(chunk_size=1, decode_unicode=False):
        yield b'DUMMY'
        raise RequestException('stream error')
        yield  # noqa

    mock_response.iter_content = iter_content

    bytes_written, hash_hex = downloader.download(
        output_file, mock_response, download_version, b2api.session
    )

    assert bytes_written == file_size
    assert output_file.getvalue() == b'DUMMY' + b'dummy' * 19


def test_download_file__data_stream_error__persistent_errors(
    apiver_module, b2api, bucket, downloader, output_file
):
    file_size = 1000
    mock_response, download_version = mock_download_response_factory(
        apiver_module, bucket, file_size=file_size
    )

    # Ensure that follow-up requests also return errors
    def iter_content(chunk_size=1, decode_unicode=False):
        yield b'd'
        raise RequestException('stream error')

    mock_response.iter_content = iter_content

    bucket.api.services.session.download_file_from_url = Mock(return_value=mock_response)

    with pytest.raises(RequestException):
        downloader.download(output_file, mock_response, download_version, b2api.session)


def test_download_file__data_stream_error__multiple_errors_recovery(
    apiver_module, b2api, bucket, downloader, output_file
):
    """Test downloader handles multiple half-failed requests and still downlaods entire file."""
    # This works since each part is attempted up to 15 times before giving up
    file_size = 100
    mock_response, download_version = mock_download_response_factory(
        apiver_module, bucket, file_size=file_size
    )

    def first_iter_content(chunk_size=1, decode_unicode=False):
        yield mock_response.raw.read(1)
        raise RequestException('stream error')

    mock_response.iter_content = first_iter_content

    download_func = bucket.api.services.session.download_file_from_url

    def download_func_mock(*args, **kwargs):
        response = download_func(*args, **kwargs).__enter__()

        def iter_content(chunk_size=1, decode_unicode=False):
            yield response.raw.read(1).upper()
            raise RequestException('stream error')

        response.iter_content = iter_content
        return response

    bucket.api.services.session.download_file_from_url = download_func_mock

    bytes_written, hash_hex = downloader.download(
        output_file, mock_response, download_version, b2api.session
    )

    assert bytes_written == file_size
    assert output_file.getvalue() == b'dUMMY' + b'DUMMY' * 19



######################################################################
#
# File: test/integration/test_download.py
#
# Copyright 2021 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import gzip
import io
import os
import pathlib
import platform
import tempfile
from pprint import pprint
from unittest import mock

import pytest

from b2sdk._internal.utils import Sha1HexDigest
from b2sdk._internal.utils.filesystem import _IS_WINDOWS
from b2sdk.v2 import *

from .base import IntegrationTestBase
from .helpers import authorize


class TestDownload(IntegrationTestBase):
    def test_large_file(self):
        bucket = self.create_bucket()
        with mock.patch.object(
            self.info, '_recommended_part_size', new=self.info.get_absolute_minimum_part_size()
        ):
            download_manager = self.b2_api.services.download_manager
            with mock.patch.object(
                download_manager,
                'strategies',
                new=[
                    ParallelDownloader(
                        min_part_size=self.info.get_absolute_minimum_part_size(),
                        min_chunk_size=download_manager.MIN_CHUNK_SIZE,
                        max_chunk_size=download_manager.MAX_CHUNK_SIZE,
                        thread_pool=download_manager._thread_pool,
                    )
                ],
            ):
                # let's check that small file downloads do not fail with these settings
                small_file_version = bucket.upload_bytes(b'0', 'a_single_char')
                with io.BytesIO() as io_:
                    bucket.download_file_by_name('a_single_char').save(io_)
                    assert io_.getvalue() == b'0'

                f, sha1 = self._file_helper(bucket)
                if small_file_version._type() != 'large':
                    # if we are here, that's not the production server!
                    assert (
                        f.download_version.content_sha1_verified
                    )  # large files don't have sha1, lets not check

                file_info = f.download_version.file_info
                assert LARGE_FILE_SHA1 in file_info
                assert file_info[LARGE_FILE_SHA1] == sha1

    def _file_helper(
        self, bucket, sha1_sum=None, bytes_to_write: int | None = None
    ) -> tuple[DownloadVersion, Sha1HexDigest]:
        bytes_to_write = bytes_to_write or int(self.info.get_absolute_minimum_part_size()) * 2 + 1
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir = pathlib.Path(temp_dir)
            source_small_file = pathlib.Path(temp_dir) / 'source_small_file'
            with open(source_small_file, 'wb') as small_file:
                self.write_zeros(small_file, bytes_to_write)
            bucket.upload_local_file(
                source_small_file,
                'small_file',
                sha1_sum=sha1_sum,
            )
            target_small_file = pathlib.Path(temp_dir) / 'target_small_file'

            f = bucket.download_file_by_name('small_file')
            f.save_to(target_small_file)

            source_sha1 = hex_sha1_of_file(source_small_file)
            assert source_sha1 == hex_sha1_of_file(target_small_file)
        return f, source_sha1

    def test_small(self):
        bucket = self.create_bucket()
        f, _ = self._file_helper(bucket, bytes_to_write=1)
        assert f.download_version.content_sha1_verified

    def test_small_unverified(self):
        bucket = self.create_bucket()
        f, _ = self._file_helper(bucket, sha1_sum='do_not_verify', bytes_to_write=1)
        if f.download_version.content_sha1_verified:
            pprint(f.download_version._get_args_for_clone())
            assert not f.download_version.content_sha1_verified


@pytest.mark.parametrize('size_multiplier', [1, 100])
def test_gzip(b2_auth_data, bucket, tmp_path, b2_api, size_multiplier):
    """Test downloading gzipped files of varius sizes with and without content-encoding."""
    source_file = tmp_path / 'compressed_file.gz'
    downloaded_uncompressed_file = tmp_path / 'downloaded_uncompressed_file'
    downloaded_compressed_file = tmp_path / 'downloaded_compressed_file'

    data_to_write = b"I'm about to be compressed and sent to the cloud, yay!\n" * size_multiplier
    source_file.write_bytes(gzip.compress(data_to_write))
    file_version = bucket.upload_local_file(
        str(source_file), 'gzipped_file', file_info={'b2-content-encoding': 'gzip'}
    )
    b2_api.download_file_by_id(file_id=file_version.id_).save_to(str(downloaded_compressed_file))
    assert downloaded_compressed_file.read_bytes() == source_file.read_bytes()

    decompressing_api, _ = authorize(b2_auth_data, B2HttpApiConfig(decode_content=True))
    decompressing_api.download_file_by_id(file_id=file_version.id_).save_to(
        str(downloaded_uncompressed_file)
    )
    assert downloaded_uncompressed_file.read_bytes() == data_to_write


@pytest.fixture
def source_file(tmp_path):
    source_file = tmp_path / 'source.txt'
    source_file.write_text('hello world')
    return source_file


@pytest.fixture
def uploaded_source_file_version(bucket, source_file):
    file_version = bucket.upload_local_file(str(source_file), source_file.name)
    return file_version


@pytest.mark.skipif(platform.system() == 'Windows', reason='no os.mkfifo() on Windows')
def test_download_to_fifo(bucket, tmp_path, source_file, uploaded_source_file_version, bg_executor):
    output_file = tmp_path / 'output.txt'
    os.mkfifo(output_file)
    output_string = None

    def reader():
        nonlocal output_string
        output_string = output_file.read_text()

    reader_future = bg_executor.submit(reader)

    bucket.download_file_by_id(file_id=uploaded_source_file_version.id_).save_to(output_file)

    reader_future.result(timeout=1)
    assert source_file.read_text() == output_string


@pytest.fixture
def binary_cap(request):
    """
    Get best suited capture.

    For Windows we need capsys as capfd fails, while on any other (i.e. POSIX systems) we need capfd.
    This is sadly tied directly to how .save_to() is implemented, as Windows required special handling.
    """
    cap = request.getfixturevalue('capsysbinary' if _IS_WINDOWS else 'capfdbinary')
    yield cap


def test_download_to_stdout(bucket, source_file, uploaded_source_file_version, binary_cap):
    output_file = 'CON' if _IS_WINDOWS else '/dev/stdout'

    bucket.download_file_by_id(file_id=uploaded_source_file_version.id_).save_to(output_file)

    assert binary_cap.readouterr().out == source_file.read_bytes()



######################################################################
#
# File: test/unit/test_progress.py
#
# Copyright 2024 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import pytest
from apiver_deps import TqdmProgressListener, make_progress_listener


@pytest.mark.parametrize(
    'tqdm_available, quiet, expected_listener',
    [
        (True, False, 'TqdmProgressListener'),
        (False, False, 'SimpleProgressListener'),
        (False, True, 'DoNothingProgressListener'),
    ],
)
def test_make_progress_listener(tqdm_available, quiet, expected_listener, monkeypatch):
    if not tqdm_available:
        monkeypatch.setattr('b2sdk._internal.progress.tqdm', None)

    assert make_progress_listener('description', quiet).__class__.__name__ == expected_listener


def test_tqdm_progress_listener__without_tqdm_module(monkeypatch):
    monkeypatch.setattr('b2sdk._internal.progress.tqdm', None)

    with pytest.raises(ModuleNotFoundError):
        TqdmProgressListener('description')



######################################################################
#
# File: test/integration/helpers.py
#
# Copyright 2022 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import os
import secrets

from b2sdk.v2 import (
    BUCKET_NAME_CHARS_UNIQ,
    BUCKET_NAME_LENGTH_RANGE,
    DEFAULT_HTTP_API_CONFIG,
    B2Api,
    InMemoryAccountInfo,
)

GENERAL_BUCKET_NAME_PREFIX = 'sdktst'
BUCKET_NAME_LENGTH = BUCKET_NAME_LENGTH_RANGE[1]
BUCKET_CREATED_AT_MILLIS = 'created_at_millis'

RNG = secrets.SystemRandom()


def _bucket_name_prefix_part(length: int) -> str:
    return ''.join(RNG.choice(BUCKET_NAME_CHARS_UNIQ) for _ in range(length))


def get_bucket_name_prefix(rnd_len: int = 8) -> str:
    return GENERAL_BUCKET_NAME_PREFIX + _bucket_name_prefix_part(rnd_len)


def random_bucket_name(prefix: str = GENERAL_BUCKET_NAME_PREFIX) -> str:
    return prefix + _bucket_name_prefix_part(BUCKET_NAME_LENGTH - len(prefix))


def authorize(b2_auth_data, api_config=DEFAULT_HTTP_API_CONFIG):
    info = InMemoryAccountInfo()
    b2_api = B2Api(info, api_config=api_config)
    realm = os.environ.get('B2_TEST_ENVIRONMENT', 'production')
    b2_api.authorize_account(realm, *b2_auth_data)
    return b2_api, info



######################################################################
#
# File: test/unit/utils/test_filesystem.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import os
import pathlib
import platform

import pytest
from apiver_deps import (
    STDOUT_FILEPATH,
    points_to_fifo,
    points_to_stdout,
)

EXPECTED_STDOUT_PATH = pathlib.Path('CON' if platform.system() == 'Windows' else '/dev/stdout')


class TestPointsToFifo:
    @pytest.mark.skipif(platform.system() == 'Windows', reason='no os.mkfifo() on Windows')
    def test_fifo_path(self, tmp_path):
        fifo_path = tmp_path / 'fifo'
        os.mkfifo(fifo_path)
        assert points_to_fifo(fifo_path) is True

    def test_non_fifo_path(self, tmp_path):
        path = tmp_path / 'subdir'
        path.mkdir(parents=True)
        assert points_to_fifo(path) is False

    def test_non_existent_path(self, tmp_path):
        path = tmp_path / 'file.txt'
        assert points_to_fifo(path) is False


class TestPointsToStdout:
    def test_stdout_path(self):
        assert points_to_stdout(EXPECTED_STDOUT_PATH) is True
        assert points_to_stdout(STDOUT_FILEPATH) is True

    def test_non_stdout_path(self, tmp_path):
        path = tmp_path / 'file.txt'
        path.touch()
        assert points_to_stdout(path) is False

    def test_non_existent_stdout_path(self, tmp_path):
        path = tmp_path / 'file.txt'
        assert points_to_stdout(path) is False



######################################################################
#
# File: test/conftest.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import concurrent.futures

import pytest


@pytest.fixture
def bg_executor():
    with concurrent.futures.ThreadPoolExecutor() as executor:
        yield executor


@pytest.fixture
def apiver_module():
    """
    b2sdk apiver module fixture.

    A compatibility function that is to be replaced by `pytest-apiver` plugin in the future.
    """
    import apiver_deps  # noqa: F401

    return apiver_deps



######################################################################
#
# File: noxfile.py
#
# Copyright 2020 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import os
import pathlib
import platform
import re
import subprocess

import nox

UPSTREAM_REPO_URL = 'git@github.com:Backblaze/b2-sdk-python.git'

# Required for PDM to use nox's virtualenvs
os.environ.update({'PDM_IGNORE_SAVED_PYTHON': '1'})

CI = os.environ.get('CI') is not None
NOX_PYTHONS = os.environ.get('NOX_PYTHONS')
_NOX_EXTRAS = os.environ.get('NOX_EXTRAS')
NOX_EXTRAS = [[]] if _NOX_EXTRAS is None else list(filter(None, [_NOX_EXTRAS.split(',')]))

PYTHON_VERSIONS = (
    [
        'pypy3.9',
        'pypy3.10',
        '3.7',
        '3.8',
        '3.9',
        '3.10',
        '3.11',
        '3.12',
        '3.13',
    ]
    if NOX_PYTHONS is None
    else NOX_PYTHONS.split(',')
)


def _detect_python_nox_id() -> str:
    major, minor, *_ = platform.python_version_tuple()
    python_nox_id = f'{major}.{minor}'
    if platform.python_implementation() == 'PyPy':
        python_nox_id = f'pypy{python_nox_id}'
    return python_nox_id


if CI and not NOX_PYTHONS:
    # this is done to allow it to work even if `nox -p` was passed to nox
    PYTHON_VERSIONS = [_detect_python_nox_id()]
    print(f'CI job mode; using provided interpreter only; PYTHON_VERSIONS={PYTHON_VERSIONS!r}')

PYTHON_DEFAULT_VERSION = PYTHON_VERSIONS[-2] if len(PYTHON_VERSIONS) > 1 else PYTHON_VERSIONS[0]

PY_PATHS = ['b2sdk', 'test', 'noxfile.py']

nox.options.reuse_existing_virtualenvs = True
nox.options.sessions = [
    'lint',
    'test',
]


def pdm_install(session: nox.Session, *args: str, dev: bool = True) -> None:
    # dev dependencies are installed by default
    prod_args = [] if dev else ['--prod']
    group_args = []
    for group in args:
        group_args.extend(['--group', group])
    session.run('pdm', 'install', *prod_args, *group_args, external=True)


def skip_coverage(python_version: str | None) -> bool:
    return (python_version or platform.python_implementation()).lower().startswith('pypy')


@nox.session(name='format', python=PYTHON_DEFAULT_VERSION)
def format_(session):
    """Lint the code and apply fixes in-place whenever possible."""
    pdm_install(session, 'lint')
    session.run('ruff', 'check', '--fix', *PY_PATHS)
    session.run('ruff', 'format', *PY_PATHS)
    # session.run(
    #     'docformatter',
    #     '--in-place',
    #     '--recursive',
    #     '--wrap-summaries=100',
    #     '--wrap-descriptions=100',
    #     *PY_PATHS,
    # )


@nox.session(python=PYTHON_DEFAULT_VERSION)
def lint(session):
    """Run linters in readonly mode."""
    # We need to install 'doc' group because liccheck needs to inspect it.
    pdm_install(session, 'doc', 'lint', 'full')
    session.run('ruff', 'check', *PY_PATHS)
    session.run('ruff', 'format', *PY_PATHS)
    # session.run(
    #     'docformatter',
    #     '--check',
    #     '--recursive',
    #     '--wrap-summaries=100',
    #     '--wrap-descriptions=100',
    #     *PY_PATHS,
    # )

    session.run('pytest', 'test/static')
    session.run('liccheck', '-s', 'pyproject.toml')

    # Check if the lockfile is up to date
    session.run('pdm', 'lock', '--check', external=True)


@nox.session(python=PYTHON_VERSIONS)
@nox.parametrize('extras', NOX_EXTRAS)
def unit(session, extras):
    """Run unit tests."""
    pdm_install(session, 'test', *extras)
    args = ['--doctest-modules', '-n', 'auto']
    if not skip_coverage(session.python):
        args += ['--cov=b2sdk', '--cov-branch', '--cov-report=xml']
    # TODO: Use session.parametrize for apiver
    session.run('pytest', '--api=v3', *args, *session.posargs, 'test/unit')
    if not skip_coverage(session.python):
        args += ['--cov-append']
    session.run('pytest', '--api=v2', *args, *session.posargs, 'test/unit')
    session.run('pytest', '--api=v1', *args, *session.posargs, 'test/unit')
    session.run('pytest', '--api=v0', *args, *session.posargs, 'test/unit')

    if not skip_coverage(session.python) and not session.posargs:
        session.notify('cover')


@nox.session(python=PYTHON_VERSIONS)
@nox.parametrize('extras', NOX_EXTRAS)
def integration(session, extras):
    """Run integration tests."""
    pdm_install(session, 'test', *extras)
    session.run('pytest', '-s', *session.posargs, 'test/integration')


@nox.session(python=PYTHON_DEFAULT_VERSION)
def cleanup_old_buckets(session):
    """Remove buckets from previous test runs."""
    pdm_install(session, 'test')
    session.run('python', '-m', 'test.integration.cleanup_buckets')


@nox.session(python=PYTHON_VERSIONS)
def test(session):
    """Run all tests."""
    if session.python:
        session.notify(f'unit-{session.python}')
        session.notify(f'integration-{session.python}')
    else:
        session.notify('unit')
        session.notify('integration')


@nox.session
def cover(session):
    """Perform coverage analysis."""
    pdm_install(session, 'test')
    session.run('coverage', 'report', '--fail-under=75', '--show-missing', '--skip-covered')
    session.run('coverage', 'erase')


@nox.session(python=PYTHON_DEFAULT_VERSION)
def build(session):
    """Build the distribution."""
    session.run('pdm', 'build', external=True)

    # Set outputs for GitHub Actions
    if CI:
        with open(os.environ['GITHUB_OUTPUT'], 'a') as github_output:
            # Path have to be specified with unix style slashes even for windows,
            # otherwise glob won't find files on windows in action-gh-release.
            print('asset_path=dist/*', file=github_output)

            version = os.environ['GITHUB_REF'].replace('refs/tags/v', '')
            print(f'version={version}', file=github_output)


@nox.session(python=PYTHON_DEFAULT_VERSION)
def doc(session):
    """Build the documentation."""
    pdm_install(session, 'doc')
    session.cd('doc')
    sphinx_args = ['-b', 'html', '-T', '-W', 'source', 'build/html']
    session.run('rm', '-rf', 'build', external=True)

    if not session.interactive:
        session.run('sphinx-build', *sphinx_args)
        session.notify('doc_cover')
    else:
        sphinx_args[-2:-2] = [
            '-E',
            '--open-browser',
            '--watch',
            '../b2sdk',
            '--ignore',
            '*.pyc',
            '--ignore',
            '*~',
        ]
        session.run('sphinx-autobuild', *sphinx_args)


@nox.session
def doc_cover(session):
    """Perform coverage analysis for the documentation."""
    pdm_install(session, 'doc')
    session.cd('doc')
    sphinx_args = ['-b', 'coverage', '-T', '-W', 'source', 'build/coverage']
    report_file = 'build/coverage/python.txt'
    session.run('sphinx-build', *sphinx_args)
    session.run('cat', report_file, external=True)

    with open('build/coverage/python.txt') as fd:
        # If there is no undocumented files, the report should have only 2 lines (header)
        if sum(1 for _ in fd) != 2:
            session.error('sphinx coverage has failed')


@nox.session(python=PYTHON_DEFAULT_VERSION)
def make_release_commit(session):
    """
    Runs `towncrier build`, commits changes, tags, all that is left to do is pushing
    """
    if session.posargs:
        version = session.posargs[0]
    else:
        session.error('Provide -- {release_version} (X.Y.Z - without leading "v")')

    if not re.match(r'^\d+\.\d+\.\d+$', version):
        session.error(
            f'Provided version="{version}". Version must be of the form X.Y.Z where '
            f'X, Y and Z are integers'
        )

    local_changes = subprocess.check_output(['git', 'diff', '--stat'])
    if local_changes:
        session.error('Uncommitted changes detected')

    current_branch = subprocess.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD']).decode()
    if current_branch != 'master':
        session.log('WARNING: releasing from a branch different than master')

    pdm_install(session, 'release')
    session.run('towncrier', 'build', '--yes', '--version', version)

    session.log(
        f'CHANGELOG updated, changes ready to commit and push\n'
        f'    git remote add upstream {UPSTREAM_REPO_URL!r} 2>/dev/null || git remote get-url upstream\n'
        f'    git commit -m "release {version}"\n'
        f'    git push upstream {current_branch}\n'
        f'Wait for a CI workflow to complete successfully, before triggering CD by pushing a tag.\n'
        f'    git tag v{version}\n'
        f'    git push upstream v{version}\n'
        f'Wait for a CD workflow to complete successfully, indicates the release is done.'
    )


def load_allowed_change_types(
    project_toml: pathlib.Path = pathlib.Path('./pyproject.toml'),
) -> set[str]:
    """
    Load the list of allowed change types from the pyproject.toml file.
    """
    import tomllib

    configuration = tomllib.loads(project_toml.read_text())
    return set(entry['directory'] for entry in configuration['tool']['towncrier']['type'])


def is_changelog_filename_valid(filename: str, allowed_change_types: set[str]) -> tuple[bool, str]:
    """
    Validates whether the given filename matches our rules.
    Provides information about why it doesn't match them.
    """
    error_reasons = []

    wanted_extension = 'md'
    try:
        description, change_type, extension = filename.rsplit('.', maxsplit=2)
    except ValueError:
        # Not enough values to unpack.
        return False, 'Doesn\'t follow the "<description>.<change_type>.md" pattern.'

    # Check whether the filename ends with .md.
    if extension != wanted_extension:
        error_reasons.append(f"Doesn't end with {wanted_extension} extension.")

    # Check whether the change type is valid.
    if change_type not in allowed_change_types:
        error_reasons.append(
            f"Change type '{change_type}' doesn't match allowed types: {allowed_change_types}."
        )

    # Check whether the description makes sense.
    try:
        int(description)
    except ValueError:
        if description[0] != '+':
            error_reasons.append("Doesn't start with a number nor a plus sign.")

    return len(error_reasons) == 0, ' / '.join(error_reasons) if error_reasons else ''


def is_changelog_entry_valid(file_content: str) -> tuple[bool, str]:
    """
    We expect the changelog entry to be a valid sentence in the English language.
    This includes, but not limits to, providing a capital letter at the start
    and the full-stop character at the end.

    Note: to do this "properly", tools like `nltk` and `spacy` should be used.
    """
    error_reasons = []

    # Check whether the first character is a capital letter.
    # Not allowing special characters nor numbers at the very start.
    if not file_content[0].isalpha() or not file_content[0].isupper():
        error_reasons.append('The first character is not a capital letter.')

    # Check if the last character is a full-stop character.
    if file_content.strip()[-1] != '.':
        error_reasons.append('The last character is not a full-stop character.')

    return len(error_reasons) == 0, ' / '.join(error_reasons) if error_reasons else ''


@nox.session(python=PYTHON_DEFAULT_VERSION)
def towncrier_check(session):
    """
    Check whether all the entries in the changelog.d follow the expected naming convention
    as well as some basic rules as to their format.
    """
    expected_non_md_files = {'.gitkeep'}
    allowed_change_types = load_allowed_change_types()

    is_error = False

    for filename in pathlib.Path('./changelog.d/').glob('*'):
        # If that's an expected file, it's all right.
        if filename.name in expected_non_md_files:
            continue

        # Check whether the file matches the expected pattern.
        is_valid, error_message = is_changelog_filename_valid(filename.name, allowed_change_types)
        if not is_valid:
            session.log(f"File {filename.name} doesn't match the expected pattern: {error_message}")
            is_error = True
            continue

        # Check whether the file isn't too big.
        if filename.lstat().st_size > 16 * 1024:
            session.log(
                f'File {filename.name} content is too big – it should be smaller than 16kB.'
            )
            is_error = True
            continue

        # Check whether the file can be loaded as UTF-8 file.
        try:
            file_content = filename.read_text(encoding='utf-8')
        except UnicodeDecodeError:
            session.log(f'File {filename.name} is not a valid UTF-8 file.')
            is_error = True
            continue

        # Check whether the content of the file is anyhow valid.
        is_valid, error_message = is_changelog_entry_valid(file_content)
        if not is_valid:
            session.log(f'File {filename.name} is not a valid changelog entry: {error_message}')
            is_error = True
            continue

    if is_error:
        session.error(
            'Found errors in the changelog.d directory. Check logs above for more information'
        )



######################################################################
#
# File: test/unit/v_all/test_constants.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import apiver_deps
import pytest


@pytest.mark.apiver(from_ver=2)
def test_public_constants():
    assert set(dir(apiver_deps)) >= {
        'ALL_CAPABILITIES',
        'B2_ACCOUNT_INFO_DEFAULT_FILE',
        'B2_ACCOUNT_INFO_ENV_VAR',
        'B2_ACCOUNT_INFO_PROFILE_FILE',
        'DEFAULT_MIN_PART_SIZE',
        'DEFAULT_RECOMMENDED_UPLOAD_PART_SIZE',
        'LARGE_FILE_SHA1',
        'LIST_FILE_NAMES_MAX_LIMIT',
        'SERVER_DEFAULT_SYNC_ENCRYPTION_SETTINGS_PROVIDER',
        'SRC_LAST_MODIFIED_MILLIS',
        'SSE_B2_AES',
        'SSE_C_KEY_ID_FILE_INFO_KEY_NAME',
        'SSE_NONE',
        'UNKNOWN_KEY_ID',
        'V',
        'VERSION',
        'XDG_CONFIG_HOME_ENV_VAR',
    }



######################################################################
#
# File: test/unit/filter/test_filter.py
#
# Copyright 2024 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import pytest
from apiver_deps import Filter

from b2sdk._internal.filter import FilterMatcher


@pytest.mark.parametrize(
    ('filters', 'expr', 'expected'),
    (
        ([], 'a', True),
        ([Filter.exclude('*')], 'something', False),
        ([Filter.include('a-*')], 'a-', True),
        ([Filter.include('a-*')], 'b-', False),
        ([Filter.exclude('*.txt')], 'a.txt', False),
        ([Filter.exclude('*.txt')], 'a.csv', True),
        ([Filter.exclude('*'), Filter.include('*.[ct]sv')], 'a.csv', True),
        ([Filter.exclude('*'), Filter.include('*.[ct]sv')], 'a.tsv', True),
        ([Filter.exclude('*'), Filter.include('*.[ct]sv')], 'a.ksv', False),
        (
            [Filter.exclude('*'), Filter.include('*.[ct]sv'), Filter.exclude('a.csv')],
            'a.csv',
            False,
        ),
        ([Filter.exclude('*'), Filter.include('*.[ct]sv'), Filter.exclude('a.csv')], 'b.csv', True),
    ),
)
def test_filter_matcher(filters, expr, expected):
    assert FilterMatcher(filters).match(expr) == expected



######################################################################
#
# File: test/unit/scan/test_folder_traversal.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import codecs
import os
import pathlib
import platform
import re
import shlex
import sys
from unittest.mock import MagicMock, patch

import pytest

from b2sdk._internal.scan.folder import LocalFolder
from b2sdk._internal.scan.policies import ScanPoliciesManager
from b2sdk._internal.scan.report import ProgressReport
from b2sdk._internal.utils import fix_windows_path_limit


class TestFolderTraversal:
    def test_flat_folder(self, tmp_path):
        # Create a directory structure below with initial scanning point at tmp_path/dir:
        # tmp_path
        # └── dir
        #     ├── file1.txt
        #     ├── file2.txt
        #     └── file3.txt

        (tmp_path / 'dir').mkdir(parents=True)

        (tmp_path / 'dir' / 'file1.txt').write_text('content1')
        (tmp_path / 'dir' / 'file2.txt').write_text('content2')
        (tmp_path / 'dir' / 'file3.txt').write_text('content3')

        folder = LocalFolder(str(tmp_path / 'dir'))
        local_paths = folder.all_files(reporter=MagicMock())
        absolute_paths = [path.absolute_path for path in list(local_paths)]

        assert absolute_paths == [
            fix_windows_path_limit(str(tmp_path / 'dir' / 'file1.txt')),
            fix_windows_path_limit(str(tmp_path / 'dir' / 'file2.txt')),
            fix_windows_path_limit(str(tmp_path / 'dir' / 'file3.txt')),
        ]

    @pytest.mark.skipif(
        platform.system() == 'Windows',
        reason="Windows doesn't allow / or \\ in filenames",
    )
    def test_invalid_name(self, tmp_path):
        # Create a directory structure below with initial scanning point at tmp_path/dir:
        # tmp_path
        # └── dir
        #     ├── file1.txt
        #     ├── subdir
        #     │   └── file2.txt
        #     ├── file\bad.txt
        #     └── file[DEL]bad.txt

        (tmp_path / 'dir' / 'subdir').mkdir(parents=True)

        (tmp_path / 'dir' / 'file1.txt').write_text('content1')
        (tmp_path / 'dir' / 'subdir' / 'file2.txt').write_text('content2')
        (tmp_path / 'dir' / 'file\\bad.txt').write_text('bad1')
        (tmp_path / 'dir' / 'file\x7fbad.txt').write_text('bad2')

        reporter = ProgressReport(sys.stdout, False)
        folder = LocalFolder(str(tmp_path / 'dir'))
        local_paths = folder.all_files(reporter=reporter)
        absolute_paths = [path.absolute_path for path in list(local_paths)]

        assert reporter.has_errors_or_warnings()
        assert isinstance(reporter.warnings, list)
        assert sorted(reporter.warnings) == [
            f"WARNING: '{tmp_path}/dir/file\\bad.txt' path contains invalid name (file names must not contain '\\'). Skipping.",
            f"WARNING: '{tmp_path}/dir/file\\x7fbad.txt' path contains invalid name (file names must not contain DEL). Skipping.",
        ]
        reporter.close()

        assert absolute_paths == [
            fix_windows_path_limit(str(tmp_path / 'dir' / 'file1.txt')),
            fix_windows_path_limit(str(tmp_path / 'dir' / 'subdir' / 'file2.txt')),
        ]

    @pytest.mark.skipif(
        platform.system() == 'Windows'
        and (platform.python_implementation() == 'PyPy' or sys.version_info >= (3, 13)),
        reason=(
            'PyPy on Windows force-decodes non-UTF-8 filenames, which makes it impossible to test this case. '
            'Python 3.13 does so similarly on Windows.'
        ),
    )
    def test_invalid_unicode_filename(self, tmp_path):
        # Create a directory structure below with initial scanning point at tmp_path/dir:
        # tmp_path
        # └── dir
        #     ├── file1.txt
        #     └── XXX (invalid utf-8 filename)

        (tmp_path / 'dir').mkdir(parents=True)
        (tmp_path / 'dir' / 'file1.txt').write_text('content1')

        foreign_encoding = 'euc_jp'
        # test sanity check
        assert (
            codecs.lookup(foreign_encoding).name != codecs.lookup(sys.getfilesystemencoding()).name
        )

        invalid_utf8_path = os.path.join(bytes(tmp_path), b'dir', 'てすと'.encode(foreign_encoding))
        try:
            with open(invalid_utf8_path, 'wb') as f:
                f.write(b'content2')
        except (OSError, UnicodeDecodeError):
            pytest.skip('Cannot create invalid UTF-8 filename on this platform')

        reporter = ProgressReport(sys.stdout, False)
        folder = LocalFolder(str(tmp_path / 'dir'))
        local_paths = folder.all_files(reporter=reporter)
        absolute_paths = [path.absolute_path for path in list(local_paths)]
        assert absolute_paths == [
            fix_windows_path_limit(str(tmp_path / 'dir' / 'file1.txt')),
        ]

        assert reporter.has_errors_or_warnings()
        assert re.match(
            r"WARNING: '.+/dir/.+' path contains invalid name "
            r'\(file name must be valid Unicode, check locale\)\. Skipping\.',
            reporter.warnings[0],
        )
        assert len(reporter.warnings) == 1

        reporter.close()

    @pytest.mark.skipif(
        platform.system() == 'Windows',
        reason="Windows doesn't allow / or \\ in filenames",
    )
    def test_invalid_directory_name(self, tmp_path):
        # Create a directory structure below with initial scanning point at tmp_path/dir:
        # tmp_path
        # └── dir
        #     ├── file1.txt
        #     └── dir\bad
        #         └── file2.txt

        (tmp_path / 'dir').mkdir(parents=True)
        (tmp_path / 'dir' / 'file1.txt').write_text('content1')
        (tmp_path / 'dir' / 'dir\\bad').mkdir(parents=True)
        (tmp_path / 'dir' / 'dir\\bad' / 'file2.txt').write_text('content2')

        reporter = ProgressReport(sys.stdout, False)
        folder = LocalFolder(str(tmp_path / 'dir'))
        local_paths = folder.all_files(reporter=reporter)
        absolute_paths = [path.absolute_path for path in list(local_paths)]

        assert reporter.has_errors_or_warnings()
        assert reporter.warnings == [
            f"WARNING: '{tmp_path}/dir/dir\\bad' path contains invalid name (file names must not contain '\\'). Skipping."
        ]
        reporter.close()

        assert absolute_paths == [
            fix_windows_path_limit(str(tmp_path / 'dir' / 'file1.txt')),
        ]

    def test_folder_with_subfolders(self, tmp_path):
        # Create a directory structure below with initial scanning point at tmp_path:
        # tmp_path
        # ├── dir1
        # │   ├── file1.txt
        # │   └── file2.txt
        # └── dir2
        #     ├── file3.txt
        #     └── file4.txt

        d1 = tmp_path / 'dir1'
        d1.mkdir()
        (d1 / 'file1.txt').write_text('content1')
        (d1 / 'file2.txt').write_text('content2')

        d2 = tmp_path / 'dir2'
        d2.mkdir()
        (d2 / 'file3.txt').write_text('content3')
        (d2 / 'file4.txt').write_text('content4')

        folder = LocalFolder(str(tmp_path))
        local_paths = folder.all_files(reporter=MagicMock())
        absolute_paths = [path.absolute_path for path in list(local_paths)]

        assert absolute_paths == [
            fix_windows_path_limit(str(d1 / 'file1.txt')),
            fix_windows_path_limit(str(d1 / 'file2.txt')),
            fix_windows_path_limit(str(d2 / 'file3.txt')),
            fix_windows_path_limit(str(d2 / 'file4.txt')),
        ]

    @pytest.mark.skipif(
        platform.system() == 'Windows' and platform.python_implementation() == 'PyPy',
        reason='Symlinks not supported on PyPy/Windows',
    )
    def test_folder_with_symlink_to_file(self, tmp_path):
        # Create a directory structure below with initial scanning point at tmp_path:
        # tmp_path
        # ├── dir
        # │   └── file.txt
        # └── symlink_file.txt -> dir/file.txt

        (tmp_path / 'dir').mkdir()

        file = tmp_path / 'dir' / 'file.txt'
        file.write_text('content')

        symlink_file = tmp_path / 'symlink_file.txt'
        symlink_file.symlink_to(file)

        folder = LocalFolder(str(tmp_path))
        local_paths = folder.all_files(reporter=MagicMock())

        absolute_paths = [path.absolute_path for path in list(local_paths)]

        assert absolute_paths == [
            fix_windows_path_limit(str(file)),
            fix_windows_path_limit(str(symlink_file)),
        ]

    @pytest.mark.skipif(
        platform.system() == 'Windows' and platform.python_implementation() == 'PyPy',
        reason='Symlinks not supported on PyPy/Windows',
    )
    @pytest.mark.timeout(5)
    def test_folder_with_circular_symlink(self, tmp_path):
        # Create a directory structure below with initial scanning point at tmp_path:
        # tmp_path
        # ├── dir
        # │   └── file.txt
        # └── symlink_dir -> dir

        (tmp_path / 'dir').mkdir()
        (tmp_path / 'dir' / 'file1.txt').write_text('content1')
        symlink_dir = tmp_path / 'dir' / 'symlink_dir'
        symlink_dir.symlink_to(tmp_path / 'dir', target_is_directory=True)

        folder = LocalFolder(str(tmp_path))

        local_paths = folder.all_files(reporter=MagicMock())
        absolute_paths = [path.absolute_path for path in list(local_paths)]

        assert absolute_paths == [
            fix_windows_path_limit(str(tmp_path / 'dir' / 'file1.txt')),
            fix_windows_path_limit(str(tmp_path / 'dir' / 'symlink_dir' / 'file1.txt')),
        ]

    @pytest.mark.skipif(
        platform.system() == 'Windows' and platform.python_implementation() == 'PyPy',
        reason='Symlinks not supported on PyPy/Windows',
    )
    @pytest.mark.timeout(5)
    def test_folder_with_symlink_to_parent(self, tmp_path):
        # Create a directory structure below with the scanning point at tmp_path/parent/child/:
        #   tmp_path
        #   ├── parent
        #   │   ├── child
        #   │   │   ├── file4.txt
        #   │   │   └── grandchild
        #   │   │       ├── file5.txt
        #   │   │       └── symlink_dir -> ../../.. (symlink to tmp_path/parent)
        #   │   └── file3.txt
        #   ├── file1.txt
        #   └── file2.txt

        (tmp_path / 'parent' / 'child' / 'grandchild').mkdir(parents=True)
        (tmp_path / 'file1.txt').write_text('content1')
        (tmp_path / 'file2.txt').write_text('content2')
        (tmp_path / 'parent' / 'file3.txt').write_text('content3')
        (tmp_path / 'parent' / 'child' / 'file4.txt').write_text('content4')
        (tmp_path / 'parent' / 'child' / 'grandchild' / 'file5.txt').write_text('content5')
        symlink_dir = tmp_path / 'parent' / 'child' / 'grandchild' / 'symlink_dir'
        symlink_dir.symlink_to(tmp_path / 'parent', target_is_directory=True)

        folder = LocalFolder(str(tmp_path / 'parent' / 'child'))

        local_paths = folder.all_files(reporter=MagicMock())
        absolute_paths = [path.absolute_path for path in list(local_paths)]

        assert absolute_paths == [
            fix_windows_path_limit(str(tmp_path / 'parent' / 'child' / 'file4.txt')),
            fix_windows_path_limit(str(tmp_path / 'parent' / 'child' / 'grandchild' / 'file5.txt')),
            fix_windows_path_limit(
                str(
                    tmp_path
                    / 'parent'
                    / 'child'
                    / 'grandchild'
                    / 'symlink_dir'
                    / 'child'
                    / 'file4.txt'
                )
            ),
            fix_windows_path_limit(
                str(
                    tmp_path
                    / 'parent'
                    / 'child'
                    / 'grandchild'
                    / 'symlink_dir'
                    / 'child'
                    / 'grandchild'
                    / 'file5.txt'
                )
            ),
            fix_windows_path_limit(
                str(tmp_path / 'parent' / 'child' / 'grandchild' / 'symlink_dir' / 'file3.txt')
            ),
        ]

    @pytest.mark.skipif(
        platform.system() == 'Windows' and platform.python_implementation() == 'PyPy',
        reason='Symlinks not supported on PyPy/Windows',
    )
    @pytest.mark.timeout(5)
    def test_root_short_loop(self, tmp_path):
        # Create a symlink to the tmp_path directory itself
        # tmp_path
        # └── tmp_path_symlink -> tmp_path

        tmp_path_symlink = tmp_path / 'tmp_path_symlink'
        tmp_path_symlink.symlink_to(tmp_path, target_is_directory=True)

        folder = LocalFolder(str(tmp_path_symlink))

        local_paths = folder.all_files(reporter=MagicMock())
        absolute_paths = [path.absolute_path for path in list(local_paths)]

        assert absolute_paths == []

    @pytest.mark.skipif(
        platform.system() == 'Windows' and platform.python_implementation() == 'PyPy',
        reason='Symlinks not supported on PyPy/Windows',
    )
    @pytest.mark.timeout(5)
    def test_root_parent_loop(self, tmp_path):
        # Create a symlink that points to the parent of the initial scanning point
        # tmp_path
        # └── start
        #     ├── file.txt
        #     └── symlink -> tmp_path

        (tmp_path / 'start').mkdir()
        (tmp_path / 'start' / 'file.txt').write_text('content')
        (tmp_path / 'start' / 'symlink').symlink_to(tmp_path, target_is_directory=True)

        folder = LocalFolder(str(tmp_path / 'start'))

        local_paths = folder.all_files(reporter=MagicMock())
        absolute_paths = [path.absolute_path for path in list(local_paths)]

        assert absolute_paths == [
            fix_windows_path_limit(str(tmp_path / 'start' / 'file.txt')),
            fix_windows_path_limit(str(tmp_path / 'start' / 'symlink' / 'start' / 'file.txt')),
        ]

    @pytest.mark.skipif(
        platform.system() == 'Windows' and platform.python_implementation() == 'PyPy',
        reason='Symlinks not supported on PyPy/Windows',
    )
    def test_symlink_that_points_deeper(self, tmp_path):
        # Create a directory structure with a symlink that points to a deeper directory
        # tmp_path
        # ├── a
        # │   └── a.txt
        # └── b
        #     ├── c
        #     │   └── c.txt
        #     └── d
        #         ├── d.txt
        #         └── e
        #             └── e.txt
        # ├── f
        # │   └── f.txt
        # └── symlink -> b/d/e

        (tmp_path / 'a').mkdir()
        (tmp_path / 'a' / 'a.txt').write_text('a')
        (tmp_path / 'b' / 'c').mkdir(parents=True)
        (tmp_path / 'b' / 'c' / 'c.txt').write_text('c')
        (tmp_path / 'b' / 'd' / 'e').mkdir(parents=True)
        (tmp_path / 'b' / 'd' / 'd.txt').write_text('d')
        (tmp_path / 'b' / 'd' / 'e' / 'e.txt').write_text('e')
        (tmp_path / 'f').mkdir()
        (tmp_path / 'f' / 'f.txt').write_text('f')
        (tmp_path / 'symlink').symlink_to(tmp_path / 'b' / 'd' / 'e', target_is_directory=True)

        folder = LocalFolder(str(tmp_path))

        local_paths = folder.all_files(reporter=MagicMock())
        absolute_paths = [path.absolute_path for path in list(local_paths)]

        assert absolute_paths == [
            fix_windows_path_limit(str(tmp_path / 'a' / 'a.txt')),
            fix_windows_path_limit(str(tmp_path / 'b' / 'c' / 'c.txt')),
            fix_windows_path_limit(str(tmp_path / 'b' / 'd' / 'd.txt')),
            fix_windows_path_limit(str(tmp_path / 'b' / 'd' / 'e' / 'e.txt')),
            fix_windows_path_limit(str(tmp_path / 'f' / 'f.txt')),
            fix_windows_path_limit(str(tmp_path / 'symlink' / 'e.txt')),
        ]

    @pytest.mark.skipif(
        platform.system() == 'Windows' and platform.python_implementation() == 'PyPy',
        reason='Symlinks not supported on PyPy/Windows',
    )
    def test_symlink_that_points_up(self, tmp_path):
        # Create a directory structure with a symlink that points to a upper directory
        # tmp_path
        # ├── a
        # │   └── a.txt
        # └── b
        #     ├── c
        #     │   └── c.txt
        #     └── d
        #         ├── d.txt
        #         └── e
        #             ├── symlink -> ../../a
        #             └── e.txt

        (tmp_path / 'a').mkdir()
        (tmp_path / 'a' / 'a.txt').write_text('a')
        (tmp_path / 'b' / 'c').mkdir(parents=True)
        (tmp_path / 'b' / 'c' / 'c.txt').write_text('c')
        (tmp_path / 'b' / 'd' / 'e').mkdir(parents=True)
        (tmp_path / 'b' / 'd' / 'd.txt').write_text('d')
        (tmp_path / 'b' / 'd' / 'e' / 'e.txt').write_text('e')
        (tmp_path / 'b' / 'd' / 'e' / 'symlink').symlink_to(
            tmp_path / 'a', target_is_directory=True
        )

        folder = LocalFolder(str(tmp_path))

        local_paths = folder.all_files(reporter=MagicMock())
        absolute_paths = [path.absolute_path for path in list(local_paths)]

        assert absolute_paths == [
            fix_windows_path_limit(str(tmp_path / 'a' / 'a.txt')),
            fix_windows_path_limit(str(tmp_path / 'b' / 'c' / 'c.txt')),
            fix_windows_path_limit(str(tmp_path / 'b' / 'd' / 'd.txt')),
            fix_windows_path_limit(str(tmp_path / 'b' / 'd' / 'e' / 'e.txt')),
            fix_windows_path_limit(str(tmp_path / 'b' / 'd' / 'e' / 'symlink' / 'a.txt')),
        ]

    @pytest.mark.skipif(
        platform.system() == 'Windows' and platform.python_implementation() == 'PyPy',
        reason='Symlinks not supported on PyPy/Windows',
    )
    @pytest.mark.timeout(5)
    def test_elaborate_infinite_loop(self, tmp_path):
        # Create a directory structure with an elaborate infinite loop of symlinks
        # tmp_path
        # ├── a
        # │   └── a.txt
        # ├── b -> c
        # ├── c -> d
        # ├── d -> e
        # ├── e -> b
        # └── f
        #     └── f.txt

        (tmp_path / 'a').mkdir()
        (tmp_path / 'a' / 'a.txt').write_text('a')
        (tmp_path / 'b').symlink_to('c')
        (tmp_path / 'c').symlink_to('d')
        (tmp_path / 'd').symlink_to('e')
        (tmp_path / 'e').symlink_to('b')
        (tmp_path / 'f').mkdir()
        (tmp_path / 'f' / 'f.txt').write_text('f')

        folder = LocalFolder(str(tmp_path))

        local_paths = folder.all_files(reporter=MagicMock())
        absolute_paths = [path.absolute_path for path in list(local_paths)]

        assert absolute_paths == [
            fix_windows_path_limit(str(tmp_path / 'a' / 'a.txt')),
            fix_windows_path_limit(str(tmp_path / 'f' / 'f.txt')),
        ]

    @pytest.mark.skipif(
        platform.system() == 'Windows' and platform.python_implementation() == 'PyPy',
        reason='Symlinks not supported on PyPy/Windows',
    )
    def test_valid_symlink_pattern_where_the_link_goes_down_and_up(self, tmp_path):
        # tmp_path
        # ├── a
        # │   └── a.txt
        # ├── b -> c/d
        # ├── c
        # │   └── d
        # │       └── b.txt
        # ├── d -> e
        # ├── e
        # │   └── e.txt
        # └── f
        #     └── f.txt

        (tmp_path / 'a').mkdir()
        (tmp_path / 'a' / 'a.txt').write_text('a')
        (tmp_path / 'b').symlink_to(tmp_path / 'c' / 'd', target_is_directory=True)
        (tmp_path / 'c').mkdir()
        (tmp_path / 'c' / 'd').mkdir()
        (tmp_path / 'c' / 'd' / 'b.txt').write_text('b')
        (tmp_path / 'd').symlink_to(tmp_path / 'e', target_is_directory=True)
        (tmp_path / 'e').mkdir()
        (tmp_path / 'e' / 'e.txt').write_text('e')
        (tmp_path / 'f').mkdir()
        (tmp_path / 'f' / 'f.txt').write_text('f')

        folder = LocalFolder(str(tmp_path))

        local_paths = folder.all_files(reporter=MagicMock())
        absolute_paths = [path.absolute_path for path in list(local_paths)]

        assert absolute_paths == [
            fix_windows_path_limit(str(tmp_path / 'a' / 'a.txt')),
            fix_windows_path_limit(str(tmp_path / 'b' / 'b.txt')),
            fix_windows_path_limit(str(tmp_path / 'c' / 'd' / 'b.txt')),
            fix_windows_path_limit(str(tmp_path / 'd' / 'e.txt')),
            fix_windows_path_limit(str(tmp_path / 'e' / 'e.txt')),
            fix_windows_path_limit(str(tmp_path / 'f' / 'f.txt')),
        ]

    @pytest.mark.skipif(
        platform.system() == 'Windows' and platform.python_implementation() == 'PyPy',
        reason='Symlinks not supported on PyPy/Windows',
    )
    def test_valid_symlink_pattern_where_the_link_goes_up_and_down(self, tmp_path):
        # Create a directory structure with a valid symlink pattern where the link goes up and down
        # tmp_path
        # ├── a
        # │   └── a.txt
        # ├── b
        # │   └── c -> ../d
        # ├── d
        # │   └── e
        # │       └── f
        # │           └── f.txt
        # └── t.txt

        (tmp_path / 'a').mkdir()
        (tmp_path / 'a' / 'a.txt').write_text('a')
        (tmp_path / 'b').mkdir()
        (tmp_path / 'b' / 'c').symlink_to(tmp_path / 'd', target_is_directory=True)
        (tmp_path / 'd').mkdir()
        (tmp_path / 'd' / 'e').mkdir()
        (tmp_path / 'd' / 'e' / 'f').mkdir()
        (tmp_path / 'd' / 'e' / 'f' / 'f.txt').write_text('f')
        (tmp_path / 't.txt').write_text('t')

        folder = LocalFolder(str(tmp_path))

        local_paths = folder.all_files(reporter=MagicMock())
        absolute_paths = [path.absolute_path for path in list(local_paths)]

        assert absolute_paths == [
            fix_windows_path_limit(str(tmp_path / 'a' / 'a.txt')),
            fix_windows_path_limit(str(tmp_path / 'b' / 'c' / 'e' / 'f' / 'f.txt')),
            fix_windows_path_limit(str(tmp_path / 'd' / 'e' / 'f' / 'f.txt')),
            fix_windows_path_limit(str(tmp_path / 't.txt')),
        ]

    @pytest.mark.skipif(
        platform.system() == 'Windows' and platform.python_implementation() == 'PyPy',
        reason='Symlinks not supported on PyPy/Windows',
    )
    @pytest.mark.timeout(5)
    def test_loop_that_goes_down_and_up(self, tmp_path):
        # Create a directory structure with a loop that goes down and up
        # tmp_path
        # ├── a
        # │   └── a.txt
        # ├── b -> c/d
        # ├── c
        # │   └── d -> ../e
        # ├── e -> b
        # └── f
        #     └── f.txt

        (tmp_path / 'a').mkdir()
        (tmp_path / 'a' / 'a.txt').write_text('a')
        (tmp_path / 'b').symlink_to(tmp_path / 'c' / 'd', target_is_directory=True)
        (tmp_path / 'c').mkdir()
        (tmp_path / 'c' / 'd').symlink_to(tmp_path / 'e', target_is_directory=True)
        (tmp_path / 'e').symlink_to('b')
        (tmp_path / 'f').mkdir()
        (tmp_path / 'f' / 'f.txt').write_text('f')

        folder = LocalFolder(str(tmp_path))

        local_paths = folder.all_files(reporter=MagicMock())
        absolute_paths = [path.absolute_path for path in list(local_paths)]

        assert absolute_paths == [
            fix_windows_path_limit(str(tmp_path / 'a' / 'a.txt')),
            fix_windows_path_limit(str(tmp_path / 'f' / 'f.txt')),
        ]

    @pytest.mark.skipif(
        platform.system() == 'Windows' and platform.python_implementation() == 'PyPy',
        reason='Symlinks not supported on PyPy/Windows',
    )
    @pytest.mark.timeout(5)
    def test_loop_that_goes_up_and_down(self, tmp_path):
        # Create a directory structure with a loop that goes up and down
        # tmp_path
        # ├── a
        # │   └── a.txt
        # ├── b
        # │   └── c -> ../d
        # ├── d
        # │   └── e
        # │       └── f -> ../../b/c
        # └── g
        #     └── g.txt

        (tmp_path / 'a').mkdir()
        (tmp_path / 'a' / 'a.txt').write_text('a')
        (tmp_path / 'b').mkdir()
        (tmp_path / 'b' / 'c').symlink_to(tmp_path / 'd', target_is_directory=True)
        (tmp_path / 'd').mkdir()
        (tmp_path / 'd' / 'e').mkdir()
        (tmp_path / 'd' / 'e' / 'f').symlink_to(tmp_path / 'b' / 'c', target_is_directory=True)
        (tmp_path / 'g').mkdir()
        (tmp_path / 'g' / 'g.txt').write_text('g')

        folder = LocalFolder(str(tmp_path))

        local_paths = folder.all_files(reporter=MagicMock())
        absolute_paths = [path.absolute_path for path in list(local_paths)]

        assert absolute_paths == [
            fix_windows_path_limit(str(tmp_path / 'a' / 'a.txt')),
            fix_windows_path_limit(str(tmp_path / 'g' / 'g.txt')),
        ]

    def test_folder_all_files__dir_excluded_by_regex(self, tmp_path):
        """
        bar$ regex should exclude bar directory and all files inside it
        """
        d1_dir = tmp_path / 'd1'
        d1_dir.mkdir()
        (d1_dir / 'file1.txt').touch()

        bar_dir = tmp_path / 'bar'
        bar_dir.mkdir()
        (bar_dir / 'file2.txt').touch()

        scan_policy = ScanPoliciesManager(exclude_dir_regexes=['bar$'])

        folder = LocalFolder(tmp_path)
        local_paths = folder.all_files(reporter=None, policies_manager=scan_policy)
        absolute_paths = [path.absolute_path for path in local_paths]

        assert absolute_paths == [
            fix_windows_path_limit(str(d1_dir / 'file1.txt')),
        ]

    def test_excluded_no_access_check(self, tmp_path):
        """Test that a directory/file is not checked for access if it is excluded."""
        # Create directories and files
        excluded_dir = tmp_path / 'excluded_dir'
        excluded_dir.mkdir()
        excluded_file = excluded_dir / 'excluded_file.txt'
        excluded_file.touch()
        included_dir = tmp_path / 'included_dir'
        included_dir.mkdir()
        (included_dir / 'excluded_file.txt').touch()

        # Setup exclusion regex that matches the excluded directory/file name
        scan_policy = ScanPoliciesManager(
            exclude_dir_regexes=[r'excluded_dir$'], exclude_file_regexes=[r'.*excluded_file.txt']
        )
        reporter = ProgressReport(sys.stdout, False)

        # Patch os.access to monitor if it is called on the excluded file
        with patch('os.access', MagicMock(return_value=True)) as mocked_access:
            folder = LocalFolder(str(tmp_path))
            list(folder.all_files(reporter=reporter, policies_manager=scan_policy))

            # Verify os.access was not called for the excluded directory/file
            mocked_access.assert_not_called()

        reporter.close()

    @pytest.mark.skipif(
        platform.system() == 'Windows',
        reason='Unix-only filesystem permissions are tested',
    )
    def test_dir_without_exec_permission(self, tmp_path, fs_perm_tool):
        """Test that a excluded directory/file without permissions emits warnings."""
        no_perm_dir = tmp_path / 'no_perm_dir'
        no_perm_dir.mkdir()
        (no_perm_dir / 'file.txt').touch()
        (no_perm_dir / 'file2.txt').touch()
        # chmod -x no_perm_dir
        no_perm_dir.chmod(0o600)

        scan_policy = ScanPoliciesManager()
        reporter = ProgressReport(sys.stdout, False)

        folder = LocalFolder(str(tmp_path))
        local_paths = folder.all_files(reporter=reporter, policies_manager=scan_policy)
        absolute_paths = [path.absolute_path for path in local_paths]
        assert not absolute_paths

        # Check that no access warnings are issued for the excluded directory/file
        assert set(reporter.warnings) == {
            f'WARNING: {tmp_path/"no_perm_dir/file.txt"} could not be accessed (no permissions to read?)',
            f'WARNING: {tmp_path/"no_perm_dir/file2.txt"} could not be accessed (no permissions to read?)',
        }

        reporter.close()

    def test_without_permissions(self, tmp_path, fs_perm_tool):
        """Test that a excluded directory/file without permissions emits warnings."""
        no_perm_dir = tmp_path / 'no_perm_dir'
        no_perm_dir.mkdir()
        (no_perm_dir / 'file.txt').touch()

        included_dir = tmp_path / 'included_dir'
        included_dir.mkdir()
        (included_dir / 'no_perm_file.txt').touch()
        (included_dir / 'included_file.txt').touch()

        # Modify directory permissions to simulate lack of access
        fs_perm_tool.deny_access(included_dir / 'no_perm_file.txt')
        fs_perm_tool.deny_access(no_perm_dir)

        scan_policy = ScanPoliciesManager()
        reporter = ProgressReport(sys.stdout, False)

        folder = LocalFolder(str(tmp_path))
        local_paths = folder.all_files(reporter=reporter, policies_manager=scan_policy)
        absolute_paths = [pathlib.Path(path.absolute_path) for path in local_paths]

        # Check that only included_dir/included_file.txt was return
        assert {path.name for path in absolute_paths} == {'included_file.txt'}

        def s(p):
            # shlex.quote works differently depending if its on windows or unix
            return shlex.quote(str(p))

        # Check that no access warnings are issued for the excluded directory/file
        assert set(reporter.warnings) == {
            f'WARNING: {s(tmp_path / "no_perm_dir")} could not be accessed (no permissions to read?)',
            f'WARNING: {s(tmp_path / "included_dir/no_perm_file.txt")} could not be accessed (no permissions to read?)',
        }

        reporter.close()

    def test_excluded_without_permissions(self, tmp_path, fs_perm_tool):
        """Test that a excluded directory/file without permissions is not processed and no warning is issued."""
        no_perm_dir = tmp_path / 'no_perm_dir'
        no_perm_dir.mkdir()
        (no_perm_dir / 'file.txt').touch()

        included_dir = tmp_path / 'included_dir'
        included_dir.mkdir()
        (included_dir / 'no_perm_file.txt').touch()
        (included_dir / 'included_file.txt').touch()

        # Modify directory permissions to simulate lack of access
        fs_perm_tool.deny_access(included_dir / 'no_perm_file.txt')
        fs_perm_tool.deny_access(no_perm_dir)

        scan_policy = ScanPoliciesManager(
            exclude_dir_regexes=[r'no_perm_dir$'], exclude_file_regexes=[r'.*no_perm_file.txt']
        )
        reporter = ProgressReport(sys.stdout, False)

        folder = LocalFolder(str(tmp_path))
        local_paths = folder.all_files(reporter=reporter, policies_manager=scan_policy)
        absolute_paths = [path.absolute_path for path in local_paths]

        # Check that only included_dir/included_file.txt was return
        assert any('included_file.txt' in path for path in absolute_paths)

        # Check that no access warnings are issued for the excluded directory/file
        assert not reporter.warnings

        reporter.close()



######################################################################
#
# File: test/integration/test_bucket.py
#
# Copyright 2024 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import pytest

from test.helpers import assert_dict_equal_ignore_extra


def test_bucket_notification_rules(bucket, b2_api):
    if 'writeBucketNotifications' not in b2_api.account_info.get_allowed()['capabilities']:
        pytest.skip('Test account does not have writeBucketNotifications capability')

    assert bucket.set_notification_rules([]) == []
    assert bucket.get_notification_rules() == []

    notification_rule = {
        'eventTypes': ['b2:ObjectCreated:*'],
        'isEnabled': True,
        'name': 'test-rule',
        'objectNamePrefix': '',
        'targetConfiguration': {
            'customHeaders': [],
            'targetType': 'webhook',
            'url': 'https://example.com/webhook',
            'hmacSha256SigningSecret': 'stringOf32AlphaNumericCharacters',
        },
    }

    set_notification_rules = bucket.set_notification_rules([notification_rule])
    assert set_notification_rules == bucket.get_notification_rules()
    assert_dict_equal_ignore_extra(
        set_notification_rules,
        [{**notification_rule, 'isSuspended': False, 'suspensionReason': ''}],
    )
    assert bucket.set_notification_rules([]) == []


def test_bucket_update__lifecycle_rules(bucket, b2_api):
    lifecycle_rule = {
        'daysFromHidingToDeleting': 1,
        'daysFromUploadingToHiding': 1,
        'daysFromStartingToCancelingUnfinishedLargeFiles': 1,
        'fileNamePrefix': '',
    }

    old_rules_list = bucket.lifecycle_rules

    updated_bucket = bucket.update(lifecycle_rules=[lifecycle_rule])
    assert updated_bucket.lifecycle_rules == [lifecycle_rule]
    assert bucket.lifecycle_rules is old_rules_list

    updated_bucket = bucket.update(lifecycle_rules=[])
    assert updated_bucket.lifecycle_rules == []



######################################################################
#
# File: test/unit/test_raw_simulator.py
#
# Copyright 2024 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import hashlib
from unittest.mock import Mock

import pytest

from b2sdk import _v3 as v3
from test.helpers import patch_bind_params


@pytest.fixture
def dummy_bucket_simulator(apiver_module):
    return apiver_module.BucketSimulator(
        Mock(spec=apiver_module.B2Api), 'account_id', 'bucket_id', 'bucket_name', 'allPublic'
    )


@pytest.fixture
def dummy_raw_simulator(apiver_module):
    return apiver_module.RawSimulator()


@pytest.fixture
def file_sim(apiver_module, dummy_bucket_simulator, file_info):
    data = b'dummy'
    return apiver_module.FileSimulator(
        account_id='account_id',
        bucket=dummy_bucket_simulator,
        file_id='dummy-id',
        action='upload',
        name='dummy.txt',
        content_type='text/plain',
        content_sha1=hashlib.sha1(data).hexdigest(),
        file_info=file_info,
        data_bytes=data,
        upload_timestamp=0,
        server_side_encryption=apiver_module.EncryptionSetting(
            mode=apiver_module.EncryptionMode.SSE_C,
            algorithm=apiver_module.EncryptionAlgorithm.AES256,
            key=apiver_module.EncryptionKey(key_id=None, secret=b'test'),
        ),
    )


def test_file_sim__as_download_headers(file_sim):
    assert file_sim.as_download_headers() == {
        'content-length': '5',
        'content-type': 'text/plain',
        'x-bz-content-sha1': '829c3804401b0727f70f73d4415e162400cbe57b',
        'x-bz-upload-timestamp': '0',
        'x-bz-file-id': 'dummy-id',
        'x-bz-file-name': 'dummy.txt',
        'X-Bz-Info-key': 'value',
        'X-Bz-Server-Side-Encryption-Customer-Algorithm': 'AES256',
        'X-Bz-Server-Side-Encryption-Customer-Key-Md5': 'CY9rzUYh03PK3k6DJie09g==',
    }


@pytest.mark.apiver(to_ver=2)
def test_bucket_simulator__upload_file__supports_file_infos(
    apiver_module, dummy_bucket_simulator, file_info
):
    """Test v2.BucketSimulator.upload_file support of deprecated file_infos param"""
    with patch_bind_params(v3.BucketSimulator, 'upload_file') as mock_method, pytest.warns(
        DeprecationWarning, match=r'deprecated argument'
    ):
        assert dummy_bucket_simulator.upload_file(
            'upload_id',
            'upload_auth_token',
            'file_name',
            123,  # content_length
            'content_type',
            'content_sha1',
            file_infos=file_info,
            data_stream='data_stream',
        )
    assert mock_method.get_bound_call_args()['file_info'] == file_info
    assert 'file_infos' not in mock_method.call_args[1]


@pytest.mark.apiver(to_ver=2)
def test_raw_simulator__get_upload_file_headers__supports_file_infos(apiver_module, file_info):
    """Test v2.RawSimulator.get_upload_file_headers support of deprecated file_infos param"""
    with patch_bind_params(v3.RawSimulator, 'get_upload_file_headers') as mock_method, pytest.warns(
        DeprecationWarning, match=r'deprecated argument'
    ):
        apiver_module.RawSimulator.get_upload_file_headers(
            upload_auth_token='upload_auth_token',
            file_name='file_name',
            content_type='content_type',
            content_length=123,
            content_sha1='content_sha1',
            server_side_encryption=None,
            file_retention=None,
            legal_hold=None,
            file_infos=file_info,
        )
    assert mock_method.get_bound_call_args()['file_info'] == file_info
    assert 'file_infos' not in mock_method.call_args[1]


@pytest.mark.apiver(to_ver=2)
def test_raw_simulator__upload_file__supports_file_infos(dummy_raw_simulator, file_info):
    """Test v2.RawSimulator.upload_file support of deprecated file_infos param"""
    with patch_bind_params(v3.RawSimulator, 'upload_file') as mock_method, pytest.warns(
        DeprecationWarning, match=r'deprecated argument'
    ):
        dummy_raw_simulator.upload_file(
            'upload_url',
            'upload_auth_token',
            'file_name',
            123,
            'content_type',
            'content_sha1',
            file_infos=file_info,
            data_stream='data_stream',
        )
    assert mock_method.get_bound_call_args()['file_info'] == file_info
    assert 'file_infos' not in mock_method.call_args[1]
